"""Module with some helpers functions to validate, parse,... parameters.

Mainly used for parameters with a complex structure that are typically reused in
different functions in geofileops.
"""

from pathlib import Path

from geofileops import LayerInfo, fileops


def validate_agg_columns(agg_columns: dict) -> None:
    """Validates if the agg_columns parameter is properly formed.

    If an problem is found, an error is raised.

    Typically used in dissolve functions.

    Args:
        agg_columns (dict): the agg_columns parameter value dict to check.

    Raises:
        ValueError: if any error is encountered.
    """
    # Check agg_columns param
    if agg_columns is None:
        return

    base_message = (
        '{"json": [<list_columns>]} '
        'or {"columns": [{"column": "...", "agg": "...", "as": "..."}, ...]}'
    )

    # It should be a dict with one key
    if (
        agg_columns is None
        or not isinstance(agg_columns, dict)
        or len(agg_columns) != 1
    ):
        message = "agg_columns must be a dict with exactly one top-level key"
        raise ValueError(f"{message}: {base_message}")

    if "json" in agg_columns:
        # The value should be a list or None
        if agg_columns["json"] is None:
            return
        if not isinstance(agg_columns["json"], list):
            message = 'agg_columns["json"] does not contain a list of strings'
            raise ValueError(f"{message}: {agg_columns['json']}: {base_message}")

        # Loop through all elements
        for agg_column in agg_columns["json"]:
            # It should be a str
            if not isinstance(agg_column, str):
                message = 'agg_columns["json"] list contains a non-string element'
                raise ValueError(f"{message}: {agg_column}: {base_message}")
    elif "columns" in agg_columns:
        supported_aggfuncs = [
            "count",
            "sum",
            "mean",
            "min",
            "max",
            "median",
            "concat",
        ]
        # The value should be a list
        if not isinstance(agg_columns["columns"], list):
            message = 'agg_columns["columns"] does not contain a list of dicts'
            raise ValueError(f"{message}: {agg_columns['columns']}: {base_message}")

        # Loop through all elements
        for agg_column in agg_columns["columns"]:
            # It should be a dict
            if not isinstance(agg_column, dict):
                message = 'agg_columns["columns"] list contains a non-dict element'
                raise ValueError(f"{message}: {agg_column}: {base_message}")

            # Check if column exists + set casing same as in data
            if "column" not in agg_column:
                message = 'each dict in agg_columns["columns"] needs a "column" element'
                raise ValueError(f"{message}: {agg_column}")
            if "agg" not in agg_column:
                message = 'each dict in agg_columns["columns"] needs an "agg" element'
                raise ValueError(f"{message}: {agg_column}")
            if "as" not in agg_column:
                message = 'each dict in agg_columns["columns"] needs an "as" element'
                raise ValueError(f"{message}: {agg_column}")
            if agg_column["agg"].lower() not in supported_aggfuncs:
                raise ValueError(
                    'agg_columns["columns"] contains unsupported aggregation '
                    f"{agg_column['agg']}, use one of {supported_aggfuncs}"
                )
            if not isinstance(agg_column["as"], str):
                raise ValueError(
                    f'agg_columns["columns"], "as" value should be string: {agg_column}'
                )
    else:
        message = (
            f"agg_columns has invalid top-level key: {next(iter(agg_columns.keys()))}"
        )
        raise ValueError(f"{message}: {base_message}")


def validate_params_single_layer(
    input_path: Path,
    output_path: Path,
    input_layer: str | LayerInfo | None,
    output_layer: str | None,
    operation_name: str,
) -> tuple[LayerInfo, str]:
    """Validate the input parameters, return the layer names.

    Args:
        input_path (Path): path to the input file
        output_path (Path): path to the output file
        input_layer (Optional[Union[str, LayerInfo]]): the layer name or the LayerInfo
            of the input file
        output_layer (Optional[str]): the layer name of the output file
        operation_name (str): the operation name, used to get clearer errors.

    Raises:
        ValueError: when an invalid parameter was passed.

    Returns:
        a tuple with the layers:
        input_layer (LayerInfo), output_layer (str)
    """
    if output_path == input_path:
        raise ValueError(f"{operation_name}: output_path must not equal input_path")
    if not input_path.exists():
        raise FileNotFoundError(f"{operation_name}: input_path not found: {input_path}")

    # Get layer info
    if not isinstance(input_layer, LayerInfo):
        input_layer = fileops.get_layerinfo(
            input_path, layer=input_layer, raise_on_nogeom=False
        )

    if output_layer is None:
        output_layer = fileops.get_default_layer(output_path)

    return input_layer, output_layer


def validate_params_two_layers(
    input1_path: Path,
    input2_path: Path | None,
    output_path: Path,
    input1_layer: str | LayerInfo | None,
    input2_layer: str | LayerInfo | None,
    output_layer: str | None,
    operation_name: str,
) -> tuple[LayerInfo, LayerInfo, str]:
    """Validate the input parameters, return the layer names.

    Args:
        input1_path (Path): path to the 1st input file
        input2_path (Path, optional): path to the 2nd input file
        output_path (Path): path to the output file
        input1_layer (Optional[Union[str, LayerInfo]]): the layer name or the LayerInfo
            of the 1st input file
        input2_layer (Optional[Union[str, LayerInfo]]): the layer name or the LayerInfo
            of the 2nd input file
        output_layer (Optional[str]): the layer name of the output file
        operation_name (str): the operation name, used to get clearer errors.

    Raises:
        ValueError: when an invalid parameter was passed.

    Returns:
        a tuple with the layers:
        input1_layer (LayerInfo), input2_layer (LayerInfo), output_layer (str)
    """
    if output_path in (input1_path, input2_path):
        raise ValueError(
            f"{operation_name}: output_path must not equal one of input paths"
        )
    if not input1_path.exists():
        raise FileNotFoundError(
            f"{operation_name}: input1_path not found: {input1_path}"
        )
    if input2_path is not None and not input2_path.exists():
        raise FileNotFoundError(
            f"{operation_name}: input2_path not found: {input2_path}"
        )

    # Get layer info
    if not isinstance(input1_layer, LayerInfo):
        input1_layer = fileops.get_layerinfo(
            input1_path, layer=input1_layer, raise_on_nogeom=False
        )
    if input2_path is not None and not isinstance(input2_layer, LayerInfo):
        input2_layer = fileops.get_layerinfo(
            input2_path, layer=input2_layer, raise_on_nogeom=False
        )

    if output_layer is None:
        output_layer = fileops.get_default_layer(output_path)

    return input1_layer, input2_layer, output_layer  # type: ignore[return-value]



"""Module containing utilities regarding processes."""

import multiprocessing
import multiprocessing.context
import os
from collections.abc import Callable
from concurrent import futures
from types import TracebackType

import psutil

WORKER_TYPES = {"threads", "processes"}


class PooledExecutorFactory:
    """Context manager to create a pooled executor.

    Args:
        worker_type (str, optional): type of executor pool to create.
            "threads" for a ThreadPoolExecutor, "processes" for a ProcessPoolExecutor.
        max_workers (int, optional): Maximum number of workers.
            Defaults to None to get automatic determination.
        initializer (function, optional): Function that does initialisations.
        mp_context (BaseContext, optional): multiprocessing context if processes are
            used. If None, "forkserver" will be used on linux to avoid risks on getting
            deadlocks. Defaults to None.

    """

    def __init__(
        self,
        worker_type: str = "processes",
        max_workers: int | None = None,
        initializer: Callable | None = None,
        initargs: tuple = (),
        mp_context: multiprocessing.context.BaseContext | None = None,
    ) -> None:
        self.worker_type = worker_type.lower()
        if self.worker_type not in WORKER_TYPES:
            raise ValueError(
                f"Invalid worker_type: {self.worker_type}. "
                f"Must be one of {WORKER_TYPES}."
            )

        self.max_workers = max_workers
        if max_workers is not None and os.name == "nt":
            # On windows, max workers should be limited to 61 to avoid errors
            self.max_workers = min(max_workers, 61)

        self.initializer = initializer
        self.initargs = initargs
        self.mp_context = mp_context
        if mp_context is None and os.name not in {"nt", "darwin"}:
            # On linux, overrule default to "forkserver" to avoid risks to deadlocks
            self.mp_context = multiprocessing.get_context("forkserver")
        self.pool: futures.Executor | None = None

    def __enter__(self) -> futures.Executor:
        if self.worker_type == "threads":
            self.pool = futures.ThreadPoolExecutor(
                max_workers=self.max_workers,
                initializer=self.initializer,
                initargs=self.initargs,
            )
        elif self.worker_type == "processes":
            self.pool = futures.ProcessPoolExecutor(
                max_workers=self.max_workers,
                initializer=self.initializer,
                initargs=self.initargs,
                mp_context=self.mp_context,
            )
        else:
            raise ValueError(
                f"Invalid worker_type: {self.worker_type}. "
                f"Must be one of {WORKER_TYPES}."
            )

        return self.pool

    def __exit__(
        self,
        type: type,  # noqa: A002
        value: Exception | None,
        traceback: TracebackType | None,
    ) -> None:
        if self.pool is not None:
            self.pool.shutdown(wait=True)


def initialize_worker(worker_type: str, nice_value: int = 15) -> None:
    """Some default inits.

    Following things are done:
    - Set the worker process priority low (to `nice_value`) so the workers don't block
      the system.
    - Reduce OpenMP threads to avoid committed memory getting very high. You can specify
      the number of threads using the environment variable `GFO_WORKER_OMP_NUM_THREADS`.

    Args:
        worker_type (str): The type of worker to initialize.
            "threads" for thread pool, "processes" for process pool.
        nice_value (int, optional): The niceness value to set for the worker. 19 is the
            maximum niceness (lowest priority), and -20 is the minimum
            (highest priority). Defaults to 15.
    """
    worker_type = worker_type.lower()
    if worker_type not in WORKER_TYPES:
        raise ValueError(
            f"Invalid worker_type: {worker_type}. Must be one of {WORKER_TYPES}."
        )

    if worker_type == "processes":
        # Reduce OpenMP threads to avoid unnecessary inflated committed memory when
        # using multiprocessing.
        # Should work for any numeric library used (openblas, mkl,...).
        # Ref: https://stackoverflow.com/questions/77764228/pandas-scipy-high-commit-memory-usage-windows
        worker_omp_threads = os.environ.get("GFO_WORKER_OMP_NUM_THREADS", "1")
        os.environ["OMP_NUM_THREADS"] = worker_omp_threads

    # We don't want the workers to block the entire system, so make the workers nice
    # if they aren't quite nice already.
    #
    # Remarks:
    #   - on linux, depending on system settings it is not possible to decrease
    #     niceness, even if it was you who niced before.
    #   - we are setting the niceness of the process here, so in case of
    #     `worker_type=threads` the main thread/process will also be impacted.
    if getprocessnice() < nice_value:
        setprocessnice(nice_value)


def getprocessnice() -> int:
    """Get the niceness of the current process.

    The nice value can (typically) range from 19, which gives all other
    processes priority, to -20, which means that this process will take
    maximum priority (which isn't very nice ;-)).

    Remarks for windows:
        - windows only supports 6 niceness classes. setprocessnice en
          getprocessnice maps niceness values to these classes.
        - when setting REALTIME priority (-20 niceness) apparently this
          results only to HIGH priority.
    """
    p = psutil.Process(os.getpid())
    nice_value = p.nice()
    if os.name == "nt":
        return process_priorityclass_to_nice(nice_value)
    else:
        return int(nice_value)


def setprocessnice(nice_value: int) -> None:
    """Set the niceness of the current process.

    The nice value can (typically) range from 19, which gives all other
    processes priority, to -20, which means that this process will take
    maximum priority (which isn't very nice ;-)).

    Remarks for windows:
        - windows only supports 6 niceness classes. setprocessnice en
          getprocessnice maps niceness values to these classes.
        - when setting REALTIME priority (-20 niceness) apparently this
          results only to HIGH priority.

    Args:
        nice_value (int): the niceness to be set.
    """
    if nice_value < -20 or nice_value > 19:
        raise ValueError(
            f"Invalid value for nice_values (min: -20, max: 19): {nice_value}"
        )
    if getprocessnice() == nice_value:
        # If the nice value is already the same... no use setting it
        return

    try:
        p = psutil.Process(os.getpid())
        if os.name == "nt":
            p.nice(process_nice_to_priorityclass(nice_value))
        else:
            p.nice(nice_value)
    except Exception as ex:  # pragma: no cover
        raise RuntimeError(
            f"Error in setprocessnice with nice_value: {nice_value}"
        ) from ex


def process_nice_to_priorityclass(nice_value: int) -> int:  # pragma: no cover
    if nice_value == -20:
        return psutil.REALTIME_PRIORITY_CLASS
    elif nice_value <= -15:
        return psutil.HIGH_PRIORITY_CLASS
    elif nice_value <= -10:
        return psutil.ABOVE_NORMAL_PRIORITY_CLASS
    elif nice_value <= 0:
        return psutil.NORMAL_PRIORITY_CLASS
    elif nice_value <= 10:
        return psutil.BELOW_NORMAL_PRIORITY_CLASS
    else:
        return psutil.IDLE_PRIORITY_CLASS


def process_priorityclass_to_nice(priority_class: int) -> int:  # pragma: no cover
    if priority_class == psutil.REALTIME_PRIORITY_CLASS:
        return -20
    elif priority_class == psutil.HIGH_PRIORITY_CLASS:
        return -15
    elif priority_class == psutil.ABOVE_NORMAL_PRIORITY_CLASS:
        return -10
    elif priority_class == psutil.NORMAL_PRIORITY_CLASS:
        return 0
    elif priority_class == psutil.BELOW_NORMAL_PRIORITY_CLASS:
        return 10
    elif priority_class == psutil.IDLE_PRIORITY_CLASS:
        return 19
    else:
        return 0



"""Module containing some general utilities."""

import datetime
import logging
import os
import re
import time
from collections.abc import Callable, Iterable
from contextlib import AbstractContextManager
from types import TracebackType
from typing import Any

logger = logging.getLogger(__name__)


class MissingRuntimeDependencyError(RuntimeError):
    """Exception raised when a geofileops runtime dependency is missing.

    Attributes:
        message (str): Exception message
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


def align_casing(string_to_align: str, strings_to_align_to: Iterable) -> str:
    """Search a string case-insentive in a list of string to align its casing.

    If the string is not found in strings_to_align_to, a ValueError is thrown.

    Args:
        string_to_align (str): string to align the casing of to strings_to_align_to.
        strings_to_align_to (Iterable): strings to align the casing with.

    Raises:
        ValueError: the string was not found in strings_to_align_to.

    Returns:
        str: the aligned string.
    """
    return align_casing_list([string_to_align], strings_to_align_to)[0]


def align_casing_list(
    strings_to_align: list[str],
    strings_to_align_to: Iterable,
    raise_on_missing: bool = True,
) -> list[str]:
    """Search the string caseintensitive in a list of strings.

    Args:
        strings_to_align (List[str]): strings to align the casing of to
            strings_to_align_to.
        strings_to_align_to (Iterable): strings to align the casing with.
        raise_on_missing (bool, optional): if True, a ValueError is raised if a string
            in ``strings_to_align`` is not found in ``strings_to_align_to``. If False,
            the casing in ``strings_to_align`` is retained in the output.

    Raises:
        ValueError: a string in strings_to_align was nog found in strings_to_align_to.

    Returns:
        List[str]: the aligned list of strings.
    """
    strings_to_align_to_upper_dict = {
        string.upper(): string for string in strings_to_align_to
    }
    strings_aligned = []
    for string in strings_to_align:
        string_aligned = strings_to_align_to_upper_dict.get(string.upper())
        if string_aligned is not None:
            strings_aligned.append(string_aligned)
        elif raise_on_missing:
            raise ValueError(f"{string} not available in: {strings_to_align_to}")
        else:
            strings_aligned.append(string)

    return strings_aligned


def report_progress(
    start_time: datetime.datetime,
    nb_done: int,
    nb_todo: int,
    operation: str | None = None,
    nb_parallel: int = 1,
) -> None:
    # If logging level not enabled for INFO, no progress reporting...
    if logger.isEnabledFor(logging.INFO) is False:
        return

    message = format_progress(
        start_time=start_time,
        nb_done=nb_done,
        nb_todo=nb_todo,
        operation=operation,
        nb_parallel=nb_parallel,
    )
    if message is not None:
        if nb_done >= nb_todo:
            message += "\n"
        print(f"\r{message}", end="", flush=True)


def format_progress(
    start_time: datetime.datetime,
    nb_done: int,
    nb_todo: int,
    operation: str | None = None,
    nb_parallel: int = 1,
) -> str | None:
    # Init
    time_passed = (datetime.datetime.now() - start_time).total_seconds()
    pct_progress = 100.0 - (nb_todo - nb_done) * 100 / nb_todo
    nb_todo_str = f"{nb_todo:n}"
    nb_decimal = len(nb_todo_str)

    # If we haven't really started yet, don't report time estimate yet
    if nb_done == 0:
        return (
            f" ?: ?: ? left, {operation} done on {nb_done:{nb_decimal}n} of "
            f"{nb_todo:{nb_decimal}n} ({pct_progress:3.2f}%)    "
        )
    else:
        pct_progress = 100.0 - (nb_todo - nb_done) * 100 / nb_todo
        if time_passed > 0:
            # Else, report progress properly...
            processed_per_hour = (nb_done / time_passed) * 3600
            # Correct the nb processed per hour if running parallel
            if nb_done < nb_parallel:
                processed_per_hour = round(processed_per_hour * nb_parallel / nb_done)
            hours_to_go = (int)((nb_todo - nb_done) / processed_per_hour)
            min_to_go = (int)((((nb_todo - nb_done) / processed_per_hour) % 1) * 60)
            secs_to_go = (int)(
                ((((nb_todo - nb_done) / processed_per_hour) % 1) * 3600) % 60
            )
            time_left_str = f"{hours_to_go:02d}:{min_to_go:02d}:{secs_to_go:02d}"
            nb_left_str = f"{nb_done:{nb_decimal}n} of {nb_todo:{nb_decimal}n}"
            pct_str = f"({pct_progress:3.2f}%)    "
        elif pct_progress >= 100:
            time_left_str = "00:00:00"
            nb_left_str = f"{nb_done:{nb_decimal}n} of {nb_todo:{nb_decimal}n}"
            pct_str = f"({pct_progress:3.2f}%)    "
        else:
            return None
        message = f"{time_left_str} left, {operation} done on {nb_left_str} {pct_str}"
        return message


def formatbytes(nb_bytes: float) -> str:
    """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
    bytes_float = float(nb_bytes)
    KB = float(1024)
    MB = float(KB**2)  # 1,048,576
    GB = float(KB**3)  # 1,073,741,824
    TB = float(KB**4)  # 1,099,511,627,776

    if bytes_float < KB:
        return "{} {}".format(bytes_float, "Bytes" if bytes_float > 1 else "Byte")
    elif KB <= bytes_float < MB:
        return f"{bytes_float / KB:.2f} KB"
    elif MB <= bytes_float < GB:
        return f"{bytes_float / MB:.2f} MB"
    elif GB <= bytes_float < TB:
        return f"{bytes_float / GB:.2f} GB"
    else:
        return f"{bytes_float / TB:.2f} TB"


def prepare_for_serialize(data: dict) -> dict:
    prepared: dict[str, Any] = {}
    for key, value in data.items():
        if isinstance(value, dict):
            prepared[key] = prepare_for_serialize(value)
        elif isinstance(value, list | tuple):
            prepared[key] = value
        else:
            prepared[key] = str(value)

    return prepared


def retry(
    max_tries: int,
    delay_fixed: float | None = None,
    delay_incremental: float | None = None,
    exceptions: tuple[type[BaseException], ...] = (Exception,),
    match: Iterable[str] | str | None = None,
) -> Callable:
    """Decorator to retry a function a number of times in case of an exception.

    Args:
        max_tries (int): maximum number of times the function is attempted.
        delay_fixed (float, optional): fixed delay in seconds between retries.
            If None, no fixed dalay is applied. Defaults to None.
        delay_incremental (float, optional): delay in seconds that increases between
            retries like this: delay = `delay_incremental` * <try_count>. If None,
            no incremental delay is applied. Defaults to None.
        exceptions (tuple[type[BaseException], ...], optional): types of exceptions to
            retry on. Defaults to (Exception,).
        match (Iterable[str] | str, optional): if specified, only exceptions with
            messages matching the given regular expression(s) are retried.
            Defaults to None.
    """
    if match is not None and isinstance(match, str):
        match = [match]

    def decorator(func: Callable) -> Callable:
        def wrapper(*args, **kwargs) -> Callable:  # noqa: ANN002, ANN003
            for try_count in range(max_tries):
                try:
                    result = func(*args, **kwargs)
                    return result
                except Exception as ex:
                    if (
                        try_count >= (max_tries - 1)
                        or not isinstance(ex, exceptions)
                        or (
                            match is not None
                            and not any(re.search(m, str(ex)) for m in match)
                        )
                    ):
                        # No more retries... raise
                        raise ex

                    logger.info(
                        f"Retrying function {func.__name__} due to exception: {ex}. "
                        f"Try {try_count + 1} of {max_tries}."
                    )
                    # Sleep if needed
                    delay = 0.0
                    if delay_fixed is not None:
                        delay = delay_fixed
                    elif delay_incremental is not None:
                        delay += delay_incremental * (try_count + 1)

                    if delay > 0:
                        time.sleep(delay)

            raise RuntimeError("Error in retry decorator. Should not be reached.")

        return wrapper

    return decorator


class TempEnv(AbstractContextManager):
    """Context manager to temporarily set/change environment variables.

    Existing values for variables are backed up and reset when the scope is left,
    variables that didn't exist before are deleted again.

    If value is None, the environment variable is deleted within the context.

    Args:
        envs (Dict[str, Any]): dict with environment variables to set.
    """

    def __init__(self, envs: dict[str, Any]) -> None:
        self._envs_backup: dict[str, str] = {}
        self._envs = envs

    def __enter__(self) -> None:
        # Only if a name and value is specified...
        for name, value in self._envs.items():
            # If the environment variable is already defined, make backup
            if name in os.environ:
                self._envs_backup[name] = os.environ[name]

            # Set env variable to value
            if value is None:
                if name in os.environ:
                    del os.environ[name]
            else:
                os.environ[name] = str(value)

    def __exit__(
        self,
        type: type[BaseException] | None,  # noqa: A002
        value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        # Set variables that were backed up back to original value
        for name, env_value in self._envs_backup.items():
            # Recover backed up value
            os.environ[name] = env_value
        # For variables without backup, remove them
        for name, _ in self._envs.items():
            if name not in self._envs_backup and name in os.environ:
                del os.environ[name]



"""General helper functions, specific for geofileops."""

import shutil
import warnings
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path

import psutil

from geofileops.helpers._options import ConfigOptions
from geofileops.util import _general_util, _io_util


@contextmanager
def create_gfo_tmp_dir(
    base_dirname: str, parent_dir: Path | None = None
) -> Iterator[Path]:
    """Context manager that creates a temporary directory for geofileops operations.

    The directory and its contents are removed when the context is exited, unless
    `ConfigOptions.remove_temp_files` is set to False.

    Args:
        base_dirname (str): The base name of the temporary directory to create. The
            following characters are replaced to "_": "/", " ".
        parent_dir (Path | None, optional): The parent directory to create the
            temporary directory in. If None, the directory specified in the environment
            variable `GFO_TMPDIR` is used. If that  that does not exist, a "geofileops"
            subdirectory in :func:`tempfile.gettempdir` is used. Defaults to None.

    Returns:
        Path: The path to the created temporary directory.
    """
    if parent_dir is None:
        parent_dir = ConfigOptions.get_tmp_dir
    base_dirname = base_dirname.replace("/", "_").replace(" ", "_")

    tmp_dir = _io_util.create_tempdir(base_dirname, parent_dir)
    try:
        yield tmp_dir
    finally:
        if ConfigOptions.get_remove_temp_files:
            shutil.rmtree(tmp_dir, ignore_errors=True)


def worker_type_to_use(input_layer_featurecount: int) -> str:
    worker_type = ConfigOptions.get_worker_type
    if worker_type in ("threads", "processes"):
        return worker_type

    # Processing in threads is 2x faster for small datasets
    if input_layer_featurecount <= 100:
        return "threads"

    return "processes"


def warn_if_low_mem(called_from: str | None = None) -> None:
    """Warning if the low memory thresshold is reached.

    Returns:
        called_from (str | None): String added in the warning that gives an indication
            where the low memory warning was triggered. Defaults to None.
    """
    min_memory_available = ConfigOptions.get_low_mem_available_warn_threshold
    virtual_available = psutil.virtual_memory().available
    if virtual_available < min_memory_available:
        virtual_available_str = _general_util.formatbytes(virtual_available)
        warning_msg = f"Low memory available: {virtual_available_str}"
        if called_from:
            warning_msg += f" ({called_from=})"
        warnings.warn(warning_msg, stacklevel=2)



"""Helper functions for working with paths to geo files."""

import os
from pathlib import Path
from typing import Any, Union

GEO_MULTI_SUFFIXES = (".gpkg.zip", ".shp.zip")


class GeoPath:
    """Class to work with geo file paths."""

    def __init__(self, path: Union[str, "os.PathLike[Any]"]) -> None:
        """Initialize the GeoPath object.

        Args:
            path (PathLike): The path to work with.
        """
        self._path = Path(path)

    @property
    def is_multi_suffix(self) -> bool:
        """Return True if the path has a geo multi-suffix.

        Returns:
            bool: True if the path has a geo multi-suffix.
        """
        name_lower = self._path.name.lower()
        for multi_suffix in GEO_MULTI_SUFFIXES:
            if name_lower.endswith(multi_suffix):
                return True

        return False

    @property
    def path(self) -> Path:
        """Return the original path as a Path object."""
        return self._path

    @property
    def name_nozip(self) -> str:
        """Return the name of a path, removing a .zip suffix if present.

        Returns:
            str: The name of the path without a .zip suffix.
        """
        if self._path.suffix.lower() == ".zip":
            return self._path.name[: -len(".zip")]
        return self._path.name

    @property
    def stem(self) -> str:
        """Return the stem of a path, supporting multi-suffixes for geo formats.

        Args:
            path (PathLike): The path to get the stem of.

        Returns:
            str: the stem of the path.
        """
        name_lower = self._path.name.lower()
        for suffixes in GEO_MULTI_SUFFIXES:
            if name_lower.endswith(suffixes):
                # Geo multi-suffix found, remove it from the name to get the stem
                return self._path.name[: -len(suffixes)]

        return self._path.stem

    @property
    def suffix_full(self) -> str:
        """Return the suffixes of a path, supporting multi-suffixes for geo formats.

        Returns:
            str: The suffixes of the path.
        """
        # If there are multiple "."s in the name, check if it is a geo multi-suffix
        name_lower = self._path.name.lower()
        for multi_suffix in GEO_MULTI_SUFFIXES:
            if name_lower.endswith(multi_suffix):
                return self._path.name[-len(multi_suffix) :]

        return self._path.suffix

    @property
    def suffix_nozip(self) -> str:
        """Return the suffixes of a path, removing a .zip suffix if present.

        Returns:
            str: The suffixes of the path without a .zip suffix.
        """
        suffix_tmp = self.suffix_full
        if suffix_tmp.lower().endswith(".zip"):
            return suffix_tmp[: -len(".zip")]

        return suffix_tmp

    def with_stem(self, stem: str) -> Path:
        """Return a Path with a new stem, supporting multi-suffixes for geo formats.

        If the input is a string, the output will be a string. If the input is a Path,
        the output will be a Path.

        Args:
            stem (str): The new stem to set.

        Returns:
            Path: The path with the new stem.

        """
        # Determine if the filename has a geo multi-suffix (case insensitive)
        new_path_str = None
        name_lower = self._path.name.lower()
        is_geo_multi_suffix = False
        for multi_suffix in GEO_MULTI_SUFFIXES:
            if name_lower.endswith(multi_suffix):
                is_geo_multi_suffix = True
                break

        # The file name has a geo multi-suffix
        if is_geo_multi_suffix:
            # Recover the suffixes from the file name to keep the casing
            multi_suffix = self._path.name[-len(multi_suffix) :]
            new_name = f"{stem}{multi_suffix}"
            # Use os.path to retain forward/backward slashes in the path
            parent, _ = os.path.split(self._path)
            new_path_str = f"{parent}/{new_name}"

        # If the input was a Path, return a Path
        if new_path_str is not None:
            return Path(new_path_str)
        else:
            return self._path.with_stem(stem)

    def with_stem_suffix(self, stem_suffix: str) -> Path:
        """Return a Path with a new stem by adding a stem_suffix.

        Multiple suffixes are supported for geo formats.

        If the input is a string, the output will be a string. If the input is a Path,
        the output will be a Path.

        Args:
            stem_suffix (str): The suffix to add to the stem.

        Returns:
            Path: The path with the new stem.

        """
        return self._path.parent / f"{self.stem}{stem_suffix}{self.suffix_full}"



"""Helper class to access the geofileops runtime options."""

import multiprocessing
import os
import tempfile
from contextlib import AbstractContextManager
from pathlib import Path
from typing import Any, Literal, Union

from pyproj import CRS


class classproperty(property):
    """Decorator to create class-level properties."""

    def __get__(self, owner_self, owner_cls):  # noqa: ANN001, ANN204
        return self.fget(owner_cls)


class _RestoreOriginalHandler(AbstractContextManager):
    """Context manager to restore the original value of an environment variable."""

    def __init__(self, key: str, original_value: str | None) -> None:
        self.key = key
        self.original_value = original_value

    def __enter__(self) -> None:
        pass

    def __exit__(self, *args: object) -> None:
        if self.original_value is None:
            if self.key in os.environ:
                del os.environ[self.key]
        else:
            os.environ[self.key] = self.original_value


class ConfigOptions:
    """Class to get and set the geofileops runtime options.

    The runtime options are saved to and read from environment variables.

    The setter methods can be used in two ways:

        1. Permanently set the option by calling the setter method directly.
        2. Temporarily set the option by using the setter method as a context manager.

    """

    @staticmethod
    def set_copy_layer_sqlite_direct(enable: bool | None) -> _RestoreOriginalHandler:
        """Enable option to copy data directly in SQLite in `copy_layer` when possible.

        If not set, this option is enabled by default.

        This can be significantly faster than having the data pass through GDAL for
        large datasets.

        It is only applied if several conditions are met:

            - only used for Geopackage files
            - only `write_mode="append"`
            - only if `explodecollections=False`
            - only if `reprojection=False`
            - ...

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_COPY_LAYER_SQLITE_DIRECT` to "TRUE" or "FALSE".

        .. versionadded:: 0.11.0

        Args:
            enable (bool | None): If True, this option is enabled. If None, the option
                is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_copy_layer_sqlite_direct(False)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_copy_layer_sqlite_direct(False):
                    gfo.copy_layer(...)

        """
        key = "GFO_COPY_LAYER_SQLITE_DIRECT"
        original_value = os.environ.get(key)
        if enable is not None:
            os.environ[key] = "TRUE" if enable else "FALSE"
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_copy_layer_sqlite_direct(cls) -> bool:
        """Should copy_layer use sqlite directly when possible.

        Returns:
            bool: True to use sqlite directly to copy layers. Defaults to True.
        """
        return _get_bool("GFO_COPY_LAYER_SQLITE_DIRECT", default=True)

    @staticmethod
    def set_io_engine(
        engine: Literal["pyogrio-arrow", "pyogrio", "fiona"] | None,
    ) -> _RestoreOriginalHandler:
        """Set the IO engine to use for reading and writing files to/from GeoDataFrames.

        Possible options are:

            - **"pyogrio-arrow"** (default if not set): use the pyogrio library via the
              arrow batch interface. The arrow batch interface will only be used if
              `pyarrow` is installed.
            - **"pyogrio"**: use the pyogrio library via the traditional interface.
            - **"fiona"**: use the fiona library. This is deprecated and support will be
              removed in a future release.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_IO_ENGINE` to one of "PYOGRIO-ARROW", "PYOGRIO", or "FIONA".

        .. versionadded:: 0.11.0

        Args:
            engine (Literal["pyogrio-arrow", "pyogrio", "fiona"] | None): The IO engine
                to use. If None, the option is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_io_engine("pyogrio")


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_io_engine("pyogrio"):
                    gfo.read_file(...)

        """
        key = "GFO_IO_ENGINE"
        original_value = os.environ.get(key)
        if engine is not None:
            os.environ[key] = engine.upper()
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_io_engine(cls) -> str:
        """The IO engine to use.

        Returns:
            str: the IO engine to use. Possible values (lowercase):

                - "pyogrio-arrow" (default if not set): use the pyogrio library via the
                  arrow batch interface.
                - "pyogrio": use the pyogrio library via the traditional interface.
                - "fiona": use the fiona library.
        """
        io_engine = (
            os.environ.get("GFO_IO_ENGINE", default="pyogrio-arrow").strip().lower()
        )
        supported_values = ["pyogrio", "fiona", "pyogrio-arrow"]
        if io_engine not in supported_values:
            raise ValueError(
                f"invalid value for configoption <GFO_IO_ENGINE>: '{io_engine}', "
                f"should be one of {supported_values}"
            )

        return io_engine

    @staticmethod
    def set_low_mem_available_warn_threshold(
        min_bytes_available: int | None,
    ) -> _RestoreOriginalHandler:
        """Set the threshold to start warning for low available memory.

        If not set, warnings are given once less than 500 MB of memory is available.

        Warnings are given for all functions that use multiprocessing, and the memory
        availability check is ran every time before a new batch is started. Note that
        this is only a warning: the operation will still be attempted even if low on
        memory. Also note that a warning is not guaranteed to be given, as memory
        usage of worker processes can increase (a lot) between checks, either due to
        geofileops processing or external factors.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_MIN_MEMORY_AVAILABLE_WARNING` to a string representing the number of
              bytes.

        .. versionadded:: 0.11.0

        Args:
            min_bytes_available (int | None): The minimum amount of memory (in bytes)
                that should be available before giving warnings. If None, the option
                is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_memory_available_warn_threshold(800 * 1024 * 1024)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_memory_available_warn_threshold(800 * 1024 * 1024):
                    gfo.dissolve(...)

        """
        key = "GFO_LOW_MEM_AVAILABLE_WARN_THRESHOLD"
        original_value = os.environ.get(key)
        if min_bytes_available is not None:
            os.environ[key] = str(min_bytes_available)
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_low_mem_available_warn_threshold(cls) -> int | None:
        """Get the threshold to start warning for low available memory.

        Returns:
            float | None: The minimum amount of memory (in bytes) that should be
                available before giving warnings. Defaults to None.
        """
        fraction_str = os.environ.get("GFO_LOW_MEM_AVAILABLE_WARN_THRESHOLD")
        if fraction_str is None:
            return 500 * 1024 * 1024  # 500 MB
        else:
            try:
                fraction = int(fraction_str)
            except ValueError as ex:
                raise ValueError(
                    "invalid value for configoption "
                    f"<GFO_LOW_MEM_AVAILABLE_WARN_THRESHOLD>: '{fraction_str}'"
                ) from ex

            return fraction

    @staticmethod
    def set_nb_parallel(nb_parallel: int | None) -> _RestoreOriginalHandler:
        """Set the preferred number of parallel workers to be used.

        If not set or set to <= 0, the number of available CPU cores will be used.
        If set to a positive value, that value is the preferred number of workers.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_NB_PARALLEL` to a string representing the number of CPU cores to use.

        .. versionadded:: 0.11.0

        Args:
            nb_parallel (int | None): The preferred number of workers. If None, the
                option is unset (so the default behavior will be used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_nb_parallel(10)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_nb_parallel(10):
                    gfo.intersection(...)

        """
        key = "GFO_NB_PARALLEL"
        original_value = os.environ.get(key)
        if nb_parallel is not None:
            os.environ[key] = str(nb_parallel)
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @staticmethod
    def get_nb_parallel(
        nb_parallel_overrule: int | None, nb_cpu_cores: int | None = None
    ) -> int:
        """Get the preferred number of workers to use for processing.

        Args:
            nb_parallel_overrule (int | None): If not None, the value specified
                overrules the configuration option.
            nb_cpu_cores (int | None): The number of CPU cores available. If None, this
                is determined automatically.

        Returns:
            int: The preferred number of workers to use for processing.
        """
        if nb_parallel_overrule is not None:
            nb_parallel = nb_parallel_overrule
        else:
            try:
                nb_parallel = int(os.environ.get("GFO_NB_PARALLEL", default="0"))
            except ValueError as ex:
                raise ValueError(
                    "invalid value for configoption <GFO_NB_PARALLEL>: "
                    f"'{os.environ.get('GFO_NB_PARALLEL')}'"
                ) from ex

        # If nb_parallel <= 0, use the number of available CPU cores
        if nb_parallel <= 0:
            if nb_cpu_cores is not None:
                nb_parallel = nb_cpu_cores
            else:
                nb_parallel = multiprocessing.cpu_count()

        return nb_parallel

    @staticmethod
    def set_on_data_error(
        action: Literal["raise", "warn"] | None,
    ) -> _RestoreOriginalHandler:
        """Set the preferred action to take when a data error occurs.

        Possible options are:

            - **"raise"** (default if not set): raise an exception.
            - **"warn"**: issue a warning and continue.

        Note that the "warn" option is only very selectively supported: in many cases,
        an exception will still be raised.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_ON_DATA_ERROR` to one of "RAISE" or "WARN".

        .. versionadded:: 0.11.0

        Args:
            action (Literal["raise", "warn"] | None): The action to take on data error.
                If None, the option is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_on_data_error("warn")


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_on_data_error("warn"):
                    gfo.read_file(...)

        """
        key = "GFO_ON_DATA_ERROR"
        original_value = os.environ.get(key)
        if action is not None:
            os.environ[key] = action.upper()
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_on_data_error(cls) -> str:
        """The preferred action when a data error occurs.

        Returns:
            str: the preferred action when a data error occurs. Possible values
                (lowercase):

                - "raise" (default if not set): raise an exception.
                - "warn": log a warning and continue.
        """
        value = os.environ.get("GFO_ON_DATA_ERROR")

        if value is None:
            return "raise"

        value_cleaned = value.strip().lower()
        supported_values = ["raise", "warn"]
        if value_cleaned not in supported_values:
            raise ValueError(
                f"invalid value for configoption <GFO_ON_DATA_ERROR>: '{value}', "
                f"should be one of {supported_values}"
            )

        return value_cleaned

    @staticmethod
    def set_remove_temp_files(enable: bool | None) -> _RestoreOriginalHandler:
        """Enable or disable removal of temporary files created during operations.

        If not set, the option is enabled by default, so temporary files are removed
        after the operation is complete.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_REMOVE_TEMP_FILES` to "TRUE" or "FALSE".

        .. versionadded:: 0.11.0

        Args:
            enable (bool | None): If True, temporary files will be removed after
                operations. If False, temporary files will be kept. If None, the option
                is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_remove_temp_files(False)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_remove_temp_files(False):
                    gfo.buffer(...)

        """
        key = "GFO_REMOVE_TEMP_FILES"
        original_value = os.environ.get(key)
        if enable is not None:
            os.environ[key] = "TRUE" if enable else "FALSE"
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_remove_temp_files(cls) -> bool:
        """Should temporary files be removed or not.

        Returns:
            bool: True to remove temp files. Defaults to True.
        """
        return _get_bool("GFO_REMOVE_TEMP_FILES", default=True)

    @staticmethod
    def set_sliver_tolerance(tolerance: float | None) -> _RestoreOriginalHandler:
        """Tolerance to filter out slivers from overlay operations between polygons.

        If 0.0, no sliver filtering is done. If negative, only slivers with tolerance
        abs(value) are retained in the output instead of filtering them out.

        If the tolerance is not set, the default depends on the ``crs``. If ``crs`` is a
        projected CRS, the default tolerance is 0.001 CRS units (typically meters or
        feet). If it is a geographic CRS, the default tolerance 1e-7 CRS units
        (typically degrees). If ``crs`` is None or invalid, the default tolerance is
        0.0, so no sliver filtering is done.

        The slivers meant here are very small, often very narrow polygons that are
        created as a side-effect of overlay operations between polygons. Due to the
        limitations of finite-precision floating point arithmetic used in such
        operations, a point that is "snapped" on a line segment, is sometimes not
        exactly located on the line. When calculating e.g. an intersection, this can
        lead to very small "sliver" polygons being created.

        Most of the time, such slivers are not desired in the output. Hence, they are
        filtered out based on certain criteria by default.

        The basic algorythm used to determine if a geometry is a sliver is to use the
        GEOS `set_precision` algorythm with a small tolerance. If the polygon becomes
        NULL, because it is smaller/narrower than the tolerance, it is considered a
        sliver. Note that this means the tolerance should not be interpreted as an
        absolute minimum width of polygons to retain, as the way the `set_precision`
        algorythm works is more complex than that.

        Because the `set_precision` algorythm is relatively costly to apply, geometries
        are first pre-filtered with a less expensive filter: the average width:

        .. code-block:: python

            average_width(geom) = 2 * area(geom) / length(geom)

        This formula is an approximation that works well for long, narrow polygons, but
        it underestimates the width for square or round polygons. Some examples:

           - **a 10 x 10 meter square**: `2 * (10 * 10)  / (4 * 10) = 400 / 40 = 5`,
             which is an underestimation, as the real average width is 10.
           - **a 1 x 100 meter rectangle**:
             `2 * (1 * 100) / (2 * (1 + 100)) = 200 / 202 = ~0.99`, which is almost
             correct as the real average width is 1.

        The average width being underestimated for some shapes means that some
        geometries are marked as slivers even if they are not. Because the
        `average_width` check is only a pre-filter, this is not a problem: such
        geometries will still be retained if they pass the more precise `set_precision`
        check.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_SLIVER_TOLERANCE` to a string representing the tolerance value.

        .. versionadded:: 0.11.0

        Args:
            tolerance (float | None): The sliver tolerance value. If None, the option is
                unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call the function:

            .. code-block:: python

                gfo.options.set_sliver_tolerance(0.01)

            If you want to temporarily change the option, you can use the function as a
            context manager:

            .. code-block:: python

                with gfo.options.set_sliver_tolerance(0.01):
                    gfo.intersection(...)

        """
        key = "GFO_SLIVER_TOLERANCE"
        original_value = os.environ.get(key)
        if tolerance is not None:
            os.environ[key] = str(tolerance)
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @staticmethod
    def get_sliver_tolerance(crs: CRS | None) -> float:
        """Tolerance to use to filter out slivers from overlay operations.

        Args:
            crs (CRS): The CRS of the geometries being processed. Used to determine
                the sliver tolerance. For projected CRSes, the tolerance is in the units
                of the CRS (e.g. meters). For geographic CRSes, the tolerance is in
                degrees.

        Returns:
            float: the sliver tolerance to be used. If 0.0, no sliver filtering should
                be done. If negative, only slivers with tolerance abs(value) should be
                retained in the output instead of filtering them out.
                If not set, the default depends on the ``crs``. If ``crs`` is a
                projected CRS, the default tolerance is 0.001 meters. If it is a
                geographic CRS, the default tolerance is 1etolerance 1e-7 degrees.
                If ``crs`` is None or invalid, the default tolerance is 0.0, so no
                sliver filtering is done.
        """
        try:
            tol_str = os.environ.get("GFO_SLIVER_TOLERANCE", None)
            if tol_str is not None:
                return float(tol_str)
            elif crs is None:
                return 0.0
            elif crs.is_projected:
                # Only found projected CRSs so far that use meters or feet, and for both
                # of those this tolerance is fine.
                return 0.001
            elif crs.is_geographic:
                return 1e-7
            else:  # pragma: no cover
                return 0.0

        except Exception as ex:
            raise ValueError(
                "invalid value for configoption <GFO_SLIVER_TOLERANCE>: "
                "should be a number"
            ) from ex

    @staticmethod
    def set_subdivide_check_parallel_fraction(
        fraction: int | None,
    ) -> _RestoreOriginalHandler:
        """For a file being checked in parallel, the fraction of features to check.

        The value set should be an integer representing the fraction of features to
        check.
        If not set, defaults to 5, resulting in 20% of the features being checked.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_SUBDIVIDE_CHECK_PARALLEL_FRACTION` to a string representing the
              fraction.

        .. versionadded:: 0.11.0

        Args:
            fraction (int | None): The fraction of features to check for subdivision.
                If None, the option is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_subdivide_check_parallel_fraction(10)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_subdivide_check_parallel_fraction(10):
                    gfo.intersection(...)

        """
        key = "GFO_SUBDIVIDE_CHECK_PARALLEL_FRACTION"
        original_value = os.environ.get(key)
        if fraction is not None:
            os.environ[key] = str(fraction)
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_subdivide_check_parallel_fraction(cls) -> int:
        """For a file being checked in parallel, the fraction of features to check.

        Returns:
            int: The fraction of features to check for subdivision. Defaults to 5.
        """
        fraction_str = os.environ.get(
            "GFO_SUBDIVIDE_CHECK_PARALLEL_FRACTION", default="5"
        )

        try:
            fraction = int(fraction_str)
        except ValueError as ex:
            raise ValueError(
                "invalid value for configoption "
                f"<GFO_SUBDIVIDE_CHECK_PARALLEL_FRACTION>: '{fraction_str}'"
            ) from ex

        return fraction

    @staticmethod
    def set_subdivide_check_parallel_rows(rows: int | None) -> _RestoreOriginalHandler:
        """For a file being checked in parallel, the number of rows to check.

        The value set should be an integer representing the minimum number of rows a
        file must have to check for subdivision in parallel.
        If not set, defaults to 500000.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_SUBDIVIDE_CHECK_PARALLEL_ROWS` to a string representing the number of
              rows.

        .. versionadded:: 0.11.0

        Args:
            rows (int | None): The minimum number of rows a file must have to check for
                subdivision in parallel. If None, the option is unset (so the default
                behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_subdivide_check_parallel_rows(1000000)


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_subdivide_check_parallel_rows(1000000):
                    gfo.intersection(...)

        """
        key = "GFO_SUBDIVIDE_CHECK_PARALLEL_ROWS"
        original_value = os.environ.get(key)
        if rows is not None:
            os.environ[key] = str(rows)
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_subdivide_check_parallel_rows(cls) -> int:
        """If a file has more rows, check if subdivide is needed in parallel.

        Returns:
            int: The minimum number of rows a file must have to check for subdivision in
                parallel. Defaults to 500000.
        """
        rows_str = os.environ.get("GFO_SUBDIVIDE_CHECK_PARALLEL_ROWS", default="500000")

        try:
            rows = int(rows_str)
        except ValueError as ex:
            raise ValueError(
                "invalid value for configoption "
                f"<GFO_SUBDIVIDE_CHECK_PARALLEL_ROWS>: '{rows_str}'"
            ) from ex

        return rows

    @staticmethod
    def set_tmp_dir(
        path: Union[str, "os.PathLike[Any]"] | None,
    ) -> _RestoreOriginalHandler:
        """Set the directory to use for temporary files created during processing.

        The value set should be a valid directory path. If not set, a subdirectory
        "geofileops" created in the system temp directory is used.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_TMPDIR` to the desired temporary directory path.

        .. versionadded:: 0.11.0

        Args:
            path (PathLike | str | None): The temporary directory path. If None, the
                option is unset (so the default behavior is used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_tmp_dir("/path/to/tmpdir")


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_tmp_dir("/path/to/tmpdir"):
                    gfo.buffer(...)

        """
        key = "GFO_TMPDIR"
        original_value = os.environ.get(key)
        if path is not None:
            if not isinstance(path, str):
                path = str(path)
            os.environ[key] = path
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_tmp_dir(cls) -> Path:
        """The directory to use for temporary files created during processing.

        Returns:
            Path: The directory to use for temporary files. Defaults to a "geofileops"
                subdirectory in the system temp directory.
        """
        tmp_dir_str = os.environ.get("GFO_TMPDIR")
        if tmp_dir_str is None:
            tmpdir = Path(tempfile.gettempdir()) / "geofileops"
        elif tmp_dir_str.strip() == "":
            raise ValueError(
                "GFO_TMPDIR='' environment variable found which is not supported."
            )
        else:
            tmpdir = Path(tmp_dir_str.strip())

        tmpdir.mkdir(parents=True, exist_ok=True)
        return tmpdir

    @staticmethod
    def set_worker_type(
        worker_type: Literal["processes", "threads", "auto"] | None,
    ) -> _RestoreOriginalHandler:
        """Set the type of worker to use for parallel processing.

        Possible options are:

            - **"processes"** (default if not set): use multiprocessing with separate
              processes.
            - **"threads"**: use multithreading within the same process.
            - **"auto"**: automatically choose the best worker type based on the
              operation being performed.

        Remarks:

            - You can also set the option temporarily by using this function as a
              context manager.
            - You can also set the option by directly setting the environment variable
              `GFO_WORKER_TYPE` to one of "processes", "threads", or "auto".

        .. versionadded:: 0.11.0

        Args:
            worker_type (Literal["processes", "threads", "auto"] | None): The type of
                worker to use. If None, the option is unset (so the default behavior is
                used).

        Examples:
            If you want to change the default value of the option in general, you can
            just call it as a function:

            .. code-block:: python

                gfo.options.set_worker_type("threads")


            If you want to temporarily change the option, you can use it as a context
            manager:

            .. code-block:: python

                with gfo.options.set_worker_type("threads"):
                    gfo.buffer(...)

        """
        key = "GFO_WORKER_TYPE"
        original_value = os.environ.get(key)
        if worker_type is not None:
            os.environ[key] = worker_type.upper()
        elif key in os.environ:
            del os.environ[key]

        return _RestoreOriginalHandler(key, original_value)

    @classproperty
    def get_worker_type(cls) -> str:
        """The type of workers to use for parallel processing.

        Returns:
            str: the type of workers to use. Possible values (lowercase):

                - "threads": use threads when processing in parallel.
                - "processes": use processes when processing in parallel.
                - "auto" (default if not specified): determine the type automatically.
        """
        worker_type = os.environ.get("GFO_WORKER_TYPE", default="auto").strip().lower()
        supported_values = ["threads", "processes", "auto"]
        if worker_type not in supported_values:
            raise ValueError(
                f"invalid value for configoption <GFO_WORKER_TYPE>: '{worker_type}', "
                f"should be one of {supported_values}"
            )

        return worker_type


def _get_bool(key: str, default: bool) -> bool:
    """Get the value for the environment variable ``key`` as a bool.

    Supported values (case insensitive):

       - True: "1", "YES", "TRUE"
       - False: "0", "NO", "FALSE"

    Args:
        key (str): the environement variable to read.
        default (bool): the value to return if the environement variable does not exist
            or if it is "".

    Raises:
        ValueError: if an invalid value is present in the environment variable.

    Returns:
        bool: True or False.
    """
    value = os.environ.get(key, default="")
    value_cleaned = value.strip().lower()

    # If the key is not defined, return default
    if value_cleaned == "":
        return default

    # Check the value
    if value_cleaned in ("1", "yes", "true"):
        return True
    elif value_cleaned in ("0", "no", "false"):
        return False
    else:
        raise ValueError(
            f"invalid value for bool configoption <{key}>: {value}, should be one of "
            "1, 0, YES, NO, TRUE, FALSE"
        )



"""Module containing utilities regarding sqlite/spatialite files."""

import datetime
import enum
import logging
import pprint
import shutil
import sqlite3
import time
import warnings
from collections.abc import Iterable
from pathlib import Path
from typing import TYPE_CHECKING, Any, Union

import shapely
from pygeoops import GeometryType
from pyproj import CRS, Transformer
from shapely.geometry import box

import geofileops as gfo
from geofileops.helpers._options import ConfigOptions
from geofileops.util import _sqlite_userdefined
from geofileops.util._general_util import MissingRuntimeDependencyError

if TYPE_CHECKING:  # pragma: no cover
    import os

logger = logging.getLogger(__name__)


class EmptyResultError(Exception):
    """Exception raised when the SQL statement disn't return any rows.

    Attributes:
        message (str): Exception message
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(self.message)


def spatialite_version_info() -> dict[str, str]:
    """Returns the versions of the spatialite modules.

    Versions returned: spatialite_version, geos_version.

    Raises:
        RuntimeError: if a runtime dependency is not available.

    Returns:
        Dict[str, str]: a dict with the version of the runtime dependencies.
    """
    test_path = Path(__file__).resolve().parent / "test.gpkg"
    conn = sqlite3.connect(test_path)
    try:
        load_spatialite(conn, enable_gpkg_mode=False)

        sql = "SELECT spatialite_version(), geos_version()"
        spatialite_version, geos_version = conn.execute(sql).fetchone()

    except MissingRuntimeDependencyError:  # pragma: no cover
        conn.rollback()
        raise
    except Exception as ex:  # pragma: no cover
        conn.rollback()
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()
        conn = None  # type: ignore[assignment]

    if not spatialite_version:  # pragma: no cover
        warnings.warn(
            "empty sqlite3 spatialite version: probably a geofileops dependency was "
            "not installed correctly: check the installation instructions in the "
            "geofileops docs.",
            stacklevel=1,
        )
    if not geos_version:  # pragma: no cover
        warnings.warn(
            "empty sqlite3 spatialite GEOS version: probably a geofileops dependency "
            "was not installed correctly: check the installation instructions in the "
            "geofileops docs.",
            stacklevel=1,
        )

    versions = {
        "spatialite_version": spatialite_version,
        "geos_version": geos_version,
    }

    return versions


class SqliteProfile(enum.Enum):
    DEFAULT = 0
    SPEED = 1


def add_gpkg_ogr_contents(
    database: Union[str, "os.PathLike[Any]", sqlite3.Connection],
    layer: str | None,
    force_update: bool = False,
) -> None:
    """Add a layer to the gpkg_ogr_contents table in a geopackage.

    If the table doesn't exist yet, it will be created.

    Args:
        database (Any): the database to add the gpkg_ogr_contents to. This can be a path
            to a file or an sqlite3.Connection object. If it is a connection, it does
            not need to have spatialite loaded and it won't be closed when returning.
        layer (str): the name of the layer to add to the gpkg_ogr_contents table.
            If None, only the table will be created.
        force_update (bool, optional): True to update the data if the layer already
            exists. Defaults to False.

    Raises:
        RuntimeError: an error occured while creating the table.
    """
    if isinstance(database, sqlite3.Connection):
        # If a connection is passed, use it
        conn = database
    else:
        conn = sqlite3.connect(database)

    sql = None
    try:
        # Create gpkg_ogr_contents table if it doesn't exist yet
        sql = """
            CREATE TABLE IF NOT EXISTS gpkg_ogr_contents(
                table_name TEXT NOT NULL PRIMARY KEY,
                feature_count INTEGER DEFAULT 0
            );
        """
        conn.execute(sql)

        if layer is None:
            conn.commit()
            return

        # Create triggers to keep the feature count up to date
        sql = f"""
            CREATE TRIGGER IF NOT EXISTS "trigger_insert_feature_count_{layer}"
                AFTER INSERT ON "{layer}"
            BEGIN
                UPDATE gpkg_ogr_contents
                    SET feature_count = feature_count + 1
                WHERE lower(table_name) = lower('{layer}');
            END;
        """
        conn.execute(sql)
        sql = f"""
            CREATE TRIGGER IF NOT EXISTS "trigger_delete_feature_count_{layer}"
                AFTER DELETE ON "{layer}"
            BEGIN
                UPDATE gpkg_ogr_contents
                    SET feature_count = feature_count - 1
                WHERE lower(table_name) = lower('{layer}');
            END;
        """
        conn.execute(sql)

        # Check if the layer already exists in the gpkg_ogr_contents table
        sql = f"""
            SELECT * FROM gpkg_ogr_contents
             WHERE lower(table_name) = '{layer.lower()}';
        """
        conn.row_factory = sqlite3.Row
        cursor = conn.execute(sql)
        contents = cursor.fetchall()
        if len(contents) > 0:
            # Layer already exists, check if we need to update it
            if force_update:
                sql = f"""
                    UPDATE gpkg_ogr_contents
                       SET feature_count = (SELECT COUNT(*) FROM "{layer}")
                     WHERE lower(table_name) = '{layer.lower()}';
                """
                conn.execute(sql)
            else:
                # Layer already exists and we don't want to update it, so skip it
                return
        else:
            # Layer doesn't exist yet, so add it to the gpkg_ogr_contents table
            sql = f"""
                INSERT INTO gpkg_ogr_contents (table_name, feature_count)
                  VALUES ('{layer}', (SELECT COUNT(*) FROM "{layer}"));
            """
            conn.execute(sql)

        conn.commit()

    except Exception as ex:  # pragma: no cover
        conn.rollback()
        raise RuntimeError(f"Error executing {sql}") from ex
    finally:
        # If no existing connection was passed, close the connection
        if not isinstance(database, sqlite3.Connection):
            conn.close()
            conn = None  # type: ignore[assignment]


def connect(
    path: Union[Path, "os.PathLike[Any]"], use_spatialite: bool = True
) -> sqlite3.Connection:
    """Connect to an existing spatialite database file.

    Spatialite is loaded by default, and if the file is a geopackage, geopackage mode
    is enabled.

    Args:
        path (PathLike): the path to the database file.
        use_spatialite (bool, optional): True to load spatialite extension.
            Defaults to True.

    Raises:
        FileNotFoundError: if the database file doesn't exist.
        RuntimeError: an error occured while connecting to the database.

    Returns:
        sqlite3.Connection: the connection to the database.
    """
    if not Path(path).exists():
        raise FileNotFoundError(f"Database file not found: {path}")

    conn = sqlite3.connect(path, detect_types=sqlite3.PARSE_DECLTYPES, uri=True)

    try:
        if use_spatialite:
            load_spatialite(conn, enable_gpkg_mode=Path(path).suffix.lower() == ".gpkg")

    except Exception:  # pragma: no cover
        conn.close()
        conn = None  # type: ignore[assignment]
        raise

    return conn


def create_new_spatialdb(
    path: Union[str, "os.PathLike[Any]"],
    crs_epsg: int | None = None,
    filetype: str | None = None,
) -> sqlite3.Connection:
    """Create a new spatialite database file and returns an open connection to it.

    Notes:
        - the bounds filled out in the gpkg_contents table will be the bounds of the crs

    Args:
        path (PathLike): the path the create the database file.
        crs_epsg (int, optional): crs to add to the crs reference layer.
            Defaults to None.
        filetype (str, optional): filetype of the spatial database to create. If
            specified, takes precendence over the suffix of the path. Possible values
            are "gpkg" and "sqlite". Defaults to None.

    Raises:
        ValueError: an invalid parameter value is passed.
        RuntimeError: an error occured while creating the database.

    Returns:
        sqlite3.Connection: the connection to the created database. Spatialite will be
            loaded on the connection.
    """
    # Check input parameters
    if filetype is not None:
        filetype = filetype.lower()
    else:
        suffix = Path(path).suffix.lower()
        if suffix == ".gpkg":
            filetype = "gpkg"
        elif suffix == ".sqlite":
            filetype = "sqlite"
        else:
            raise ValueError(
                f"Unsupported {suffix=} for {path=}, change suffix or specify filetype"
            )

    if crs_epsg is not None and not isinstance(crs_epsg, int):
        raise ValueError(f"Invalid {crs_epsg=}")

    # Connecting to non existing database file will create it...
    conn = sqlite3.connect(path)
    sql = None
    try:
        # Load spatialite extension. Enabling geopackage mode is only possible after the
        # typical GPKG metadata tables are created.
        load_spatialite(conn, enable_gpkg_mode=False)

        # Starting transaction manually for good performance, mainly needed on Windows.
        sql = "BEGIN TRANSACTION;"
        conn.execute(sql)

        if filetype == "gpkg":
            # Remark: this only works on the main database!
            sql = "SELECT gpkgCreateBaseTables();"
            conn.execute(sql)

            if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                sql = f"SELECT gpkgInsertEpsgSRID({crs_epsg});"
                conn.execute(sql)

            # The GPKG created till now is of version 1.0. Apply some upgrades to
            # make it 1.4.

            # Upgrade GPKG from version 1.0 to 1.4.
            # Most changes are related to rtree index triggers, but they
            # are not applicable here as this is an empty database at this point.
            # The 1.3 changes are needed: remove following metadata triggers as they
            # gave issues in some circumstances.
            # https://github.com/opengeospatial/geopackage/pull/240
            triggers_to_remove = [
                "gpkg_metadata_md_scope_insert",
                "gpkg_metadata_md_scope_update",
                "gpkg_metadata_reference_reference_scope_insert",
                "gpkg_metadata_reference_reference_scope_update",
                "gpkg_metadata_reference_column_name_insert",
                "gpkg_metadata_reference_column_name_update",
                "gpkg_metadata_reference_row_id_value_insert",
                "gpkg_metadata_reference_row_id_value_update",
                "gpkg_metadata_reference_timestamp_insert",
                "gpkg_metadata_reference_timestamp_update",
            ]
            for trigger in triggers_to_remove:
                sql = f"DROP TRIGGER IF EXISTS {trigger};"
                conn.execute(sql)

            # Set GPKG version to 1.4
            sql = "PRAGMA application_id=1196444487;"
            conn.execute(sql)
            sql = "PRAGMA user_version=10400;"
            conn.execute(sql)

            # Now we can enable geopackage mode
            sql = "SELECT EnableGpkgMode();"
            conn.execute(sql)

        elif filetype == "sqlite":
            sql = "SELECT InitSpatialMetaData(1);"
            conn.execute(sql)
            if crs_epsg is not None and crs_epsg not in [0, -1, 4326]:
                sql = f"SELECT InsertEpsgSrid({crs_epsg});"
                conn.execute(sql)

        else:
            raise ValueError(f"Unsupported {filetype=}")

        conn.commit()

    except ValueError:
        conn.close()
        conn = None  # type: ignore[assignment]
        raise
    except Exception as ex:  # pragma: no cover
        conn.close()
        conn = None  # type: ignore[assignment]
        raise RuntimeError(f"Error creating spatial db {path} executing {sql}") from ex

    return conn


def get_column_types(database: Path, table: str) -> dict[str, str]:
    """Get the column types of a table in a sqlite database.

    Args:
        database (Path): the path to the sqlite database.
        table (str): the name of the table to get the column types from.

    Raises:
        RuntimeError: an error occured while fetching the column types.

    Returns:
        dict[str, str]: a dict with the column names as keys and the column types as
            values.
    """
    column_types = {}

    # Connect to the input database and fetch the column types
    conn = sqlite3.connect(database)
    try:
        # Get column types
        sql = f"PRAGMA table_info('{table}');"
        cur = conn.execute(sql)
        for row in cur.fetchall():
            column_types[row[1]] = row[2]
        cur.close()

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error {ex} executing {sql}") from ex

    finally:
        conn.close()
        conn = None  # type: ignore[assignment]

    return column_types


def get_columns(
    sql_stmt: str,
    input_databases: dict[str, Path],
    empty_output_ok: bool = True,
    use_spatialite: bool = True,
    output_geometrytype: GeometryType | None = None,
) -> dict[str, str]:
    # Init
    start = time.perf_counter()
    tmp_dir = None

    def get_filetype(path: Path) -> str:
        if path.suffix.lower() == ".gpkg":
            return "gpkg"
        return path.suffix.lstrip(".")

    # Connect to/create sqlite main database
    conn = None
    if "main" in input_databases:
        # If an input database is main, use it as the main database
        main_db_path = input_databases["main"]
        filetype = get_filetype(main_db_path)
        conn = connect(main_db_path, use_spatialite=use_spatialite)
    else:
        # Create temp output db to be sure the output DB is writable, even though we
        # only create a temporary table.
        filetype = get_filetype(next(iter(input_databases.values())))
        conn = create_new_spatialdb(":memory:", filetype=filetype)

    sql = None
    try:
        # Attach to all input databases
        for dbname, path in input_databases.items():
            # main is already opened, so skip it
            if dbname == "main":
                continue

            sql = f"ATTACH DATABASE ? AS {dbname}"
            dbSpec = (str(path),)
            conn.execute(sql, dbSpec)

        # Set some default performance options
        database_names = ["main", *list(input_databases.keys())]
        set_performance_options(conn, SqliteProfile.SPEED, database_names)

        # Start transaction manually needed for performance
        sql = "BEGIN TRANSACTION;"
        conn.execute(sql)

        # Prepare sql statement for execute
        sql_stmt_prepared = sql_stmt.format(batch_filter="")

        # Log explain plan if debug logging enabled.
        if logger.isEnabledFor(logging.DEBUG):
            sql = f"""
                EXPLAIN QUERY PLAN
                 SELECT * FROM (
                   {sql_stmt_prepared}
                );
            """
            cur = conn.execute(sql)
            plan = cur.fetchall()
            cur.close()
            logger.debug(pprint.pformat(plan))

        # Create temp table to get the column names + general data types
        # + fetch one row to use it to determine geometrytype.
        # Remark: specify redundant OFFSET 0 to keep sqlite from flattings the subquery.
        sql = f"""
            CREATE TEMPORARY TABLE tmp AS
            SELECT *
              FROM (
                {sql_stmt_prepared}
              )
             LIMIT 1 OFFSET 0;
        """
        conn.execute(sql)
        conn.commit()

        sql = "PRAGMA TABLE_INFO(tmp)"
        cur = conn.execute(sql)
        tmpcolumns = cur.fetchall()
        cur.close()

        # Fetch one row to try to get more detailed data types if needed
        sql = "SELECT * FROM tmp"
        tmpdata = conn.execute(sql).fetchone()
        if tmpdata is not None and len(tmpdata) == 0:
            tmpdata = None
        if not empty_output_ok and tmpdata is None:
            # If no row was returned, stop
            raise EmptyResultError(f"Query didn't return any rows: {sql_stmt}")

        # Loop over all columns to determine the data type
        columns = {}
        for column_index, column in enumerate(tmpcolumns):
            columnname = column[1]
            columntype = column[2]

            if columnname == "geom":
                # PRAGMA TABLE_INFO gives None as column type for a
                # geometry column. So if output_geometrytype not specified,
                # Use ST_GeometryType to get the type
                # based on the data + apply to_multitype to be sure
                if output_geometrytype is None:
                    sql = f'SELECT ST_GeometryType("{columnname}") FROM tmp;'
                    result = conn.execute(sql).fetchall()
                    if len(result) > 0 and result[0][0] is not None:
                        # Some geometry types returned contain spaces, e.g. "POLYGON Z".
                        geometrytype = result[0][0].replace(" ", "")
                        output_geometrytype = GeometryType[geometrytype].to_multitype
                    else:
                        output_geometrytype = GeometryType["GEOMETRY"]
                columns[columnname] = output_geometrytype.name
            elif columntype == "INT":
                columns[columnname] = "INTEGER"
            elif columntype is None or columntype == "":
                sql = f'SELECT typeof("{columnname}") FROM tmp;'
                result = conn.execute(sql).fetchall()
                if len(result) > 0 and result[0][0] is not None:
                    columns[columnname] = result[0][0]
                else:
                    # If unknown, take the most general types
                    columns[columnname] = "NUMERIC"
            elif columntype == "NUM":
                # PRAGMA TABLE_INFO sometimes returns 'NUM', but apparently this
                # cannot be used in "CREATE TABLE".
                if tmpdata is None:
                    columns[columnname] = "NUMERIC"
                elif isinstance(tmpdata[column_index], datetime.date):
                    columns[columnname] = "DATE"
                elif isinstance(tmpdata[column_index], datetime.datetime):
                    columns[columnname] = "DATETIME"
                elif isinstance(tmpdata[column_index], str):
                    sql = f'SELECT datetime("{columnname}") FROM tmp;'
                    result = conn.execute(sql).fetchall()
                    if len(result) > 0 and result[0][0] is not None:
                        columns[columnname] = "DATETIME"
                    else:
                        columns[columnname] = "NUMERIC"
                else:
                    columns[columnname] = "NUMERIC"
            else:
                columns[columnname] = columntype

    except Exception as ex:  # pragma: no cover
        conn.rollback()
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()
        conn = None
        if ConfigOptions.get_remove_temp_files and tmp_dir is not None:
            shutil.rmtree(tmp_dir, ignore_errors=True)

    time_taken = time.perf_counter() - start
    if time_taken > 5:  # pragma: no cover
        logger.info(f"get_columns ready, took {time_taken:.2f} seconds")

    return columns


def copy_table(
    input_path: Union[str, "os.PathLike[Any]"],
    output_path: Union[str, "os.PathLike[Any]"],
    input_table: str,
    output_table: str,
    columns: Iterable[str] | None = None,
    where: str | None = None,
    preserve_fid: bool = False,
    profile: SqliteProfile = SqliteProfile.DEFAULT,
) -> None:
    """Copy data from one to another table.

    Notes:
        - At the moment only appending to an existing table is supported.
        - At the moment only copying from one sqlite file to another is supported.

    Args:
        input_path (PathLike): The path to the input SQLite database.
        output_path (PathLike): The path to the output SQLite database.
        input_table (str): The name of the input table.
        output_table (str): The name of the output table.
        columns (Iterable[str] | None, optional): The list of columns to copy. If None,
            all columns will be copied.
        where (str | None, optional): An optional SQL WHERE clause to filter the rows
            to copy. Defaults to None.
        preserve_fid (bool, optional): Whether to preserve values in the input FID
            column. Defaults to False.
        profile (SqliteProfile, optional): The SQLite profile to use.
            Defaults to SqliteProfile.DEFAULT.
    """
    # copy_table only supports local paths... so we can just use Path
    input_path = Path(input_path)
    output_path = Path(output_path)
    if not input_path.exists():
        raise FileNotFoundError(f"Input file not found: {input_path}")
    if not output_path.exists():
        raise FileNotFoundError(f"Output file not found: {output_path}")
    if input_path.resolve() == output_path.resolve():
        raise ValueError(f"Input and output paths cannot be the same: {input_path}")

    # Connect with spatialite loaded, as get_total_bounds needs spatialite
    conn = connect(output_path, use_spatialite=True)

    # Execute the insert statement
    is_gpkg = Path(output_path).suffix.lower() == ".gpkg"
    sql = None
    try:
        # Attach the input database
        sql = "ATTACH DATABASE ? AS input_db"
        dbSpec = (str(input_path),)
        conn.execute(sql, dbSpec)

        # Set some default performance options
        database_names = ["main", "input_db"]
        set_performance_options(conn, profile, database_names)

        # First determine the total_bounds of the result based on the total_bounds of
        # the input and output tables.
        if is_gpkg:
            # Determine total bounds of input and output tables
            input_info = get_gpkg_content(conn, input_table, db_name="input_db")
            input_bounds = input_info["total_bounds"]
            if input_bounds is None:
                input_bounds = get_gpkg_total_bounds(
                    conn, input_table, db_name="input_db"
                )
            output_info = get_gpkg_content(conn, output_table, db_name="main")
            output_bounds = output_info["total_bounds"]
            if output_bounds is None:
                output_bounds = get_gpkg_total_bounds(
                    conn, output_table, db_name="main"
                )

            # Determine the resulting bounds
            bounds_list = [
                box(*b) for b in [input_bounds, output_bounds] if b is not None
            ]
            result_bounds = (
                shapely.MultiPolygon(bounds_list).bounds
                if len(bounds_list) > 0
                else None
            )

        # Start transaction manually needed for performance
        sql = "BEGIN TRANSACTION;"
        conn.execute(sql)

        if columns is None:
            # If the columns are not specified, determine them from the input table.
            # If the input layer has fewer columns than the output, those columns will
            # simply get the default values...
            # If the fid should not be preserved, don't include it in the column list
            column_filter = "" if preserve_fid else "WHERE lower(name) <> 'fid'"
            sql = f"""
                SELECT name
                  FROM pragma_table_info('{input_table}', 'input_db')
                 {column_filter};
            """
            columns = [value[0] for value in conn.execute(sql).fetchall()]

        elif preserve_fid and "fid" not in [col.lower() for col in columns]:
            # If preserve_fid is asked, the fid should be in the list of columns.
            columns = list(columns)
            columns.append("fid")

        columns_str = ", ".join([f'"{col}"' for col in columns])
        where_clause = f"WHERE {where}" if where else ""
        sql = f"""
            INSERT INTO main."{output_table}"
                ({columns_str})
            SELECT {columns_str}
              FROM input_db."{input_table}"
             {where_clause};
        """
        conn.execute(sql)

        # Make sure the bounds in gpkg_contents are up to date for a ".gpkg" file
        if is_gpkg:
            # Update the gpkg_contents table if the bounds have changed
            if result_bounds and result_bounds != input_info["total_bounds"]:
                sql = f"""
                    UPDATE gpkg_contents
                       SET min_x = {result_bounds[0]}
                          ,min_y = {result_bounds[1]}
                          ,max_x = {result_bounds[2]}
                          ,max_y = {result_bounds[3]}
                     WHERE lower(table_name) = '{output_table.lower()}';
                """
                conn.execute(sql)

        conn.commit()
    except Exception as ex:  # pragma: no cover
        conn.rollback()

        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        conn.close()
        conn = None  # type: ignore[assignment]


def create_table_as_sql(
    input_databases: dict[str, Path],
    output_path: Path,
    sql_stmt: str,
    output_layer: str,
    output_geometrytype: GeometryType | None,
    output_crs: int,
    append: bool = False,
    update: bool = False,
    create_spatial_index: bool = False,
    create_ogr_contents: bool = False,
    empty_output_ok: bool = True,
    column_datatypes: dict | None = None,
    profile: SqliteProfile = SqliteProfile.DEFAULT,
) -> None:
    """Execute sql statement and save the result in the output file.

    Args:
        input_databases (dict): dict with the database name(s) and path(s) to the input
            database(s).
        output_path (Path): the path where the output file needs to be created/appended.
        sql_stmt (str): SELECT statement to run on the input files.
        output_layer (str): layer/table name to use.
        output_geometrytype (Optional[GeometryType]): geometry type of the output.
        output_crs (int): epsg code of crs of the output file.
        append (bool, optional): True to append to an existing file. Defaults to False.
        update (bool, optional): True to append to an existing layer. Defaults to False.
        create_spatial_index (bool, optional): True to create a spatial index on the
            output layer. Defaults to False.
        create_ogr_contents (bool, optional): True to create the gpkg_ogr_contents table
            in the output file and fill it up. Defaults to False.
        empty_output_ok (bool, optional): If the sql_stmt doesn't return any rows and
            True, create an empty output file. If False, throw EmptyResultError.
            Defaults to True.
        column_datatypes (dict, optional): Can be used to specify the data types of
            columns in the form of {"columnname": "datatype"}. If the data type of
            (some) columns is not specified, it it automatically determined as good as
            possible. Defaults to None.
        profile (SqliteProfile, optional): the set of PRAGMA's to use when creating the
            table. SqliteProfile.DEFAULT will use default setting. SqliteProfile.SPEED
            uses settings optimized for speed, but will be less save regarding
            transaction safety,...
            Defaults to SqliteProfile.DEFAULT.

    Raises:
        ValueError: invalid (combinations of) parameters passed.
        EmptyResultError: the sql_stmt didn't return any rows.
    """
    # Check input parameters
    if append is True or update is True:
        raise ValueError("append=True nor update=True are implemented.")

    # All input files and the output file must have the same suffix.
    output_suffix_lower = output_path.suffix.lower()
    for path in input_databases.values():
        if output_suffix_lower != path.suffix.lower():
            raise ValueError(
                "output_path and all input paths must have the same suffix!"
            )

    # Determine columns/datatypes to create the table if not specified
    column_types = column_datatypes
    if column_types is None:
        column_types = get_columns(
            sql_stmt=sql_stmt,
            input_databases=input_databases,
            empty_output_ok=empty_output_ok,
            use_spatialite=True,
            output_geometrytype=output_geometrytype,
        )

    # Create or open output database
    conn = None
    if not output_path.exists():
        # Output file doesn't exist yet: create and init it
        conn = create_new_spatialdb(path=output_path, crs_epsg=output_crs)
    else:
        # Output file exists: open it with spatialite loaded
        conn = connect(output_path, use_spatialite=True)

    sql = None
    try:

        def to_string_for_sql(value: object) -> str:
            if value is None:
                return "NULL"
            else:
                return str(value)

        # Connect to output database file so it is main, otherwise the
        # gpkg... functions don't work
        # Remark: sql statements using knn only work if they are main, so they
        # are executed with ogr, as the output needs to be main as well :-(.
        output_databasename = "main"

        # Attach to all input databases
        for dbname, path in input_databases.items():
            sql = f"ATTACH DATABASE ? AS {dbname}"
            dbSpec = (str(path),)
            conn.execute(sql, dbSpec)

        # Set some default performance options
        database_names = [output_databasename, *list(input_databases.keys())]
        set_performance_options(conn, profile, database_names)

        # Start transaction manually needed for performance
        sql = "BEGIN TRANSACTION;"
        conn.execute(sql)

        # If geometry type was not specified, look for it in column_types
        if output_geometrytype is None and "geom" in column_types:
            output_geometrytype = GeometryType(column_types["geom"])

        # Now we can create the table
        # Create output table using the gpkgAddGeometryColumn() function
        # Problem: the spatialite function gpkgAddGeometryColumn() doesn't support
        # layer names with special characters (eg. '-')...
        # Solution: mimic the behaviour of gpkgAddGeometryColumn manually.
        # Create table without geom column
        """
        columns_for_create = [
            f'"{columnname}" {column_types[columnname]}\n' for columnname
            in column_types if columnname != 'geom'
        ]
        sql = (
            f'CREATE TABLE {output_databasename}."{output_layer}" '
            f'({", ".join(columns_for_create)})'
        )
        conn.execute(sql)
        # Add geom column with gpkgAddGeometryColumn()
        # Remark: output_geometrytype.name should be detemined from data if needed,
        # see above...
        sql = (
            f"SELECT gpkgAddGeometryColumn("
            f"    '{output_layer}', 'geom', '{output_geometrytype.name}', 0, 0, "
            f"    {to_string_for_sql(crs_epsg)});"
        conn.execute(sql)
        """
        columns_for_create = [
            f'"{columnname}" {column_types[columnname]}\n'
            for columnname in column_types
            if columnname.lower() != "fid"
        ]
        sql = f"""
            CREATE TABLE {output_databasename}."{output_layer}" (
                fid INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
                {", ".join(columns_for_create)}
            )
        """
        conn.execute(sql)

        # Add metadata
        if output_suffix_lower == ".gpkg":
            data_type = "features" if "geom" in column_types else "attributes"

            # Fill out the bounds of the layer using the bounds of the crs if possible
            try:
                crs = CRS.from_user_input(output_crs)
                if crs is not None and crs.area_of_use is not None:
                    transformer = Transformer.from_crs(
                        crs.geodetic_crs, crs, always_xy=True
                    )
                    bounds = transformer.transform_bounds(*crs.area_of_use.bounds)
                    min_x, min_y, max_x, max_y = (
                        to_string_for_sql(coord) for coord in bounds
                    )
                else:
                    min_x = min_y = max_x = max_y = "NULL"

            except Exception:
                min_x = min_y = max_x = max_y = "NULL"

            # ~ mimic behaviour of gpkgAddGeometryColumn()
            sql = f"""
                INSERT INTO {output_databasename}.gpkg_contents (
                  table_name, data_type, identifier, description, last_change,
                  min_x, min_y, max_x, max_y, srs_id)
                VALUES ('{output_layer}', '{data_type}', NULL, '', DATETIME(),
                  {min_x}, {min_y}, {max_x}, {max_y}, {to_string_for_sql(output_crs)});
            """
            conn.execute(sql)

            # If there is a geometry column, register it
            if "geom" in column_types:
                assert output_geometrytype is not None
                sql = f"""
                    INSERT INTO {output_databasename}.gpkg_geometry_columns (
                        table_name, column_name, geometry_type_name, srs_id, z, m)
                    VALUES ('{output_layer}', 'geom', '{output_geometrytype.name}',
                        {to_string_for_sql(output_crs)}, 0, 0);
                """
                conn.execute(sql)

                # Geometry triggers were removed from the GPKG specs in 1.2!
                # Remark: this only works on the main database!
                # sql = f"SELECT gpkgAddGeometryTriggers('{output_layer}', 'geom');"
                # conn.execute(sql)

        elif output_suffix_lower == ".sqlite":
            # Create geom metadata if there is one
            if "geom" in column_types:
                assert output_geometrytype is not None
                sql = f"""
                    SELECT RecoverGeometryColumn(
                        '{output_layer}', 'geom',
                        {to_string_for_sql(output_crs)}, '{output_geometrytype.name}');
                """
                conn.execute(sql)

        # Insert data using the sql statement specified
        try:
            columns_for_insert = [f'"{column}"' for column in column_types]
            sql = (
                f'INSERT INTO {output_databasename}."{output_layer}" '
                f"({', '.join(columns_for_insert)})\n{sql_stmt}"
            )
            conn.execute(sql)

        except Exception as ex:
            ex_message = str(ex).lower()
            if ex_message.startswith(
                "unique constraint failed:"
            ) and ex_message.endswith(".fid"):
                ex.args = (
                    f"{ex}: avoid this by not selecting or aliasing fid "
                    '("select * will select fid!)',
                )
            raise

        # Fill out the feature count in the gpkg_ogr_contents table + create triggers
        if create_ogr_contents and output_suffix_lower == ".gpkg":
            add_gpkg_ogr_contents(database=conn, layer=output_layer, force_update=True)

        # Create spatial index if needed
        if create_spatial_index and "geom" in column_types:
            sql = f"SELECT UpdateLayerStatistics('{output_layer}', 'geom');"
            conn.execute(sql)
            if output_suffix_lower == ".gpkg":
                warnings.warn(
                    "Using create_spatial_index=True for a GPKG file is not "
                    "recommended as it is slow and creates outdated triggers; "
                    "better use gfo.create_spatial_index() afterwards.",
                    stacklevel=2,
                )
                # Create the necessary empty index, triggers,...
                sql = f"SELECT gpkgAddSpatialIndex('{output_layer}', 'geom');"
                conn.execute(sql)
                # Now fill the index
                sql = f"""
                    INSERT INTO "rtree_{output_layer}_geom"
                      SELECT fid
                            ,ST_MinX(geom), ST_MaxX(geom), ST_MinY(geom), ST_MaxY(geom)
                        FROM "{output_layer}"
                       WHERE geom IS NOT NULL
                         AND ST_IsEmpty(geom) = 0
                """
                conn.execute(sql)
            elif output_suffix_lower == ".sqlite":
                sql = f"SELECT CreateSpatialIndex('{output_layer}', 'geom');"
                conn.execute(sql)

        conn.commit()

    except EmptyResultError:
        logger.info(f"Query didn't return any rows: {sql_stmt}")
        conn.close()
        conn = None
        if output_path.exists():
            output_path.unlink()
    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error {ex} executing {sql}") from ex
    finally:
        if conn is not None:
            conn.close()
            conn = None


def execute_sql(
    path: Path, sql_stmt: str | list[str], use_spatialite: bool = True
) -> None:
    # Connect to database file
    conn = connect(path, use_spatialite=use_spatialite)
    sql = None

    try:
        """
        # Set nb KB of cache
        sql = "PRAGMA cache_size=-50000;"
        conn.execute(sql)
        sql = "PRAGMA temp_store=MEMORY;"
        conn.execute(sql)
        """
        if isinstance(sql_stmt, str):
            sql = sql_stmt
            conn.execute(sql)
        else:
            for sql in sql_stmt:
                conn.execute(sql)
        conn.commit()

    except Exception as ex:
        conn.rollback()
        raise RuntimeError(f"Error executing {sql}") from ex
    finally:
        conn.close()


def get_gpkg_content(
    database: Union[Path, "os.PathLike[Any]", sqlite3.Connection],
    table_name: str,
    db_name: str = "main",
) -> dict[str, Any]:
    """Get the content info of a specific table of a geopackage.

    This function retrieves the following metadata from the `gpkg_contents` table for
    the specified table:

        - table_name: Name of the table/layer.
        - data_type: Type of data (e.g., features, tiles, etc.).
        - identifier: Unique identifier for the layer.
        - description: Description of the layer.
        - last_change: Timestamp of the last change to the layer.
        - min_x: Minimum X coordinate of the layer's total bounding box.
        - min_y: Minimum Y coordinate of the layer's total bounding box.
        - max_x: Maximum X coordinate of the layer's total bounding box.
        - max_y: Maximum Y coordinate of the layer's total bounding box.
        - total_bounds: The total bounds of the layer. If the layer does not contain any
          non-empty geometries, the value is None. Otherwise, this is a list [min_x,
          min_y, max_x, max_y].
        - srs_id: Spatial Reference System Identifier.

    Args:
        database (PathLike or Connection): file path or connection to the geopackage.
            If a connection is passed, it is not closed by this function.
        table_name (str): name of the table to get the contents for.
        db_name (str, optional): name of the database to use in case of
            an attached database. Defaults to "main".

    Returns:
        dict[str, Any]: the information for the specified table.

    """
    if isinstance(database, sqlite3.Connection):
        # If a connection is passed, use it
        conn = database
    else:
        # Otherwise create a new connection, but we don't need spatialite here
        conn = sqlite3.connect(database)

    sql = None
    try:
        sql = f"""
            SELECT table_name, data_type, identifier, description, last_change,
                   min_x, min_y, max_x, max_y, srs_id
              FROM {db_name}.gpkg_contents
             WHERE lower(table_name) = '{table_name.lower()}';
        """
        conn.row_factory = sqlite3.Row
        cursor = conn.execute(sql)
        contents = cursor.fetchall()[0]
        info: dict[str, object | None] = dict(contents)
        bounds = [info["min_x"], info["min_y"], info["max_x"], info["max_y"]]
        info["total_bounds"] = bounds if all(b is not None for b in bounds) else None

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex

    finally:
        # If no existing connection was passed, close the connection
        if not isinstance(database, sqlite3.Connection):
            conn.close()
            conn = None  # type: ignore[assignment]

    return info


def get_gpkg_contents(
    database: Union[Path, "os.PathLike[Any]", sqlite3.Connection],
    database_name: str = "main",
) -> dict[str, dict]:
    """Get the contents of the gpkg_contents table of a geopackage.

    Args:
        database (PathLike or Connection): file path or connection to the geopackage.
        database_name (str, optional): name of the database to use in case of
            an attached database. Defaults to "main".

    Returns:
        dict[str, Any]: the contents of the geopackage.

    """
    if isinstance(database, sqlite3.Connection):
        # If a connection is passed, use it
        conn = database
    else:
        # Otherwise create a new connection, but we don't need spatialite here
        conn = sqlite3.connect(database)

    sql = None
    try:
        sql = f"""
            SELECT table_name, data_type, identifier, description, last_change,
                   min_x, min_y, max_x, max_y, srs_id
              FROM {database_name}.gpkg_contents;
        """
        conn.row_factory = sqlite3.Row
        cursor = conn.execute(sql)
        contents = cursor.fetchall()
        contents_dict = {row["table_name"]: dict(row) for row in contents}

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex

    finally:
        # If no existing connection was passed, close the connection
        if not isinstance(database, sqlite3.Connection):
            conn.close()
            conn = None  # type: ignore[assignment]

    return contents_dict


def get_gpkg_geometry_column_info(
    database: Union[Path, "os.PathLike[Any]", sqlite3.Connection],
    table_name: str,
    db_name: str = "main",
) -> dict[str, Any]:
    """Get information about the geometry column of a specific table in a geopackage.

    This function retrieves the following metadata from the `gpkg_geometry_columns`
    table for the specified table:
        - column_name: Name of the geometry column.
        - geometry_type_name: Type of geometry (e.g., POINT, LINESTRING, POLYGON).
        - srs_id: Spatial Reference System Identifier.
        - z: Whether the geometry has Z (elevation) values.
        - m: Whether the geometry has M (measure) values.

    Args:
        database (PathLike or Connection): file path or connection to the geopackage.
        table_name (str): name of the table to get the geometry columns for.
        db_name (str, optional): name of the database to use in case of
            an attached database. Defaults to "main".

    Returns:
        dict[str, Any]: the geometry columns for the specified table.

    """
    if isinstance(database, sqlite3.Connection):
        # If a connection is passed, use it
        conn = database
    else:
        # Otherwise create a new connection, but we don't need spatialite here
        conn = sqlite3.connect(database)

    sql = None
    try:
        sql = f"""
            SELECT column_name, geometry_type_name, srs_id, z, m
              FROM {db_name}.gpkg_geometry_columns
             WHERE lower(table_name) = '{table_name.lower()}';
        """
        conn.row_factory = sqlite3.Row
        cursor = conn.execute(sql)
        column_info = dict(cursor.fetchone())

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex

    finally:
        # If no existing connection was passed, close the connection
        if not isinstance(database, sqlite3.Connection):
            conn.close()
            conn = None  # type: ignore[assignment]

    return column_info


def get_gpkg_ogr_contents(path: Path) -> dict[str, dict]:
    """Get the contents of the gpkg_ogr_contents table of a geopackage.

    Args:
        path (Path): file path to the geopackage.

    Returns:
        dict[str, Any]: the contents of the geopackage.
    """
    # Connect to database file, we don't need spatialite here
    conn = sqlite3.connect(path)

    sql = None
    try:
        sql = "SELECT * FROM gpkg_ogr_contents;"
        conn.row_factory = sqlite3.Row
        cursor = conn.execute(sql)
        contents = cursor.fetchall()
        contents_dict = {row["table_name"]: dict(row) for row in contents}
    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex
    finally:
        conn.close()
        conn = None  # type: ignore[assignment]

    return contents_dict


def get_tables(path: Path) -> list[str]:
    """List all tables in the database.

    Args:
        path (Path): file path to the database.

    Returns:
        list[str]: the list of all tables in the database.
    """
    # Connect to database file, we don't need spatialite here
    conn = sqlite3.connect(path)

    sql = None
    try:
        sql = "SELECT name FROM sqlite_master WHERE type='table';"
        cursor = conn.execute(sql)
        tables = [row[0] for row in cursor.fetchall()]
    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex
    finally:
        conn.close()
        conn = None  # type: ignore[assignment]

    return tables


def get_gpkg_total_bounds(
    database: Union[Path, "os.PathLike[Any]", sqlite3.Connection],
    table_name: str,
    geometry_column: str | None = None,
    db_name: str = "main",
) -> tuple[float, float, float, float] | None:
    """Get the total bounds of all layers in the geopackage.

    Args:
        database (PathLike): file path to the geopackage.
        table_name (str | None, optional): the table to get the total bounds of.
        geometry_column (str, optional): name of the geometry column.
            Defaults to "geom".
        db_name (str, optional): name of the database to use in case of
            an attached database. Defaults to "main".

    Returns:
        tuple[float, float, float, float]: the total bounds as (minx, miny, maxx, maxy)
            or None if the table does not contain any non-empty geometries.
    """
    sql = None
    try:
        if isinstance(database, sqlite3.Connection):
            # If a connection is passed, use it
            conn = database
        else:
            # Connect with spatialite loaded
            conn = connect(database, use_spatialite=True)

        if geometry_column is None:
            # If geometry column is not specified, get it from gpkg_geometry_columns
            geometry_column_info = get_gpkg_geometry_column_info(
                conn, table_name, db_name
            )
            geometry_column = geometry_column_info["column_name"]

        sql = f"""
            SELECT ST_AsBinary(Extent({geometry_column})) AS bounds
              FROM "{db_name}"."{table_name}";
        """
        cursor = conn.execute(sql)
        bounds_wkb = cursor.fetchone()[0]
        bounds = shapely.from_wkb(bounds_wkb)
        if bounds is not None and not bounds.is_empty:
            total_bounds = bounds.bounds
        else:
            total_bounds = None

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex

    finally:
        # If no existing connection was passed, close the connection
        if not isinstance(database, sqlite3.Connection):
            conn.close()
            conn = None  # type: ignore[assignment]

    return total_bounds


def test_data_integrity(path: Path, use_spatialite: bool = True) -> None:
    # Get list of layers in database
    layers = gfo.listlayers(path=path)

    # Connect to database file, with spatialite if needed
    conn = connect(path, use_spatialite=use_spatialite)
    sql = None

    try:
        # Set some basic default performance options
        set_performance_options(conn)

        # Loop over all layers to check if all data is readable
        for layer in layers:
            sql = f'SELECT * FROM "{layer}"'
            cursor = conn.execute(sql)

            # Fetch the data in chunks to avoid excessive memory usage
            while True:
                result = cursor.fetchmany(10000)
                if not result:
                    # All data was fetched from layer
                    break

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}") from ex
    finally:
        conn.close()
        conn = None  # type: ignore[assignment]


def set_performance_options(
    conn: sqlite3.Connection,
    profile: SqliteProfile | None = None,
    database_names: list[str] | None = None,
) -> None:
    """Set some performance related PRAGMA's on the sqlite connection."""
    database_names = database_names or []
    try:
        # Set cache size to 128 MB (in kibibytes)
        sql = "PRAGMA cache_size=-128000;"
        conn.execute(sql)
        # Set temp storage to MEMORY
        sql = "PRAGMA temp_store=2;"
        conn.execute(sql)
        # Set soft heap limit to 1 GB (in bytes)
        sql = f"PRAGMA soft_heap_limit={1024 * 1024 * 1024};"
        conn.execute(sql)

        # Use the sqlite profile specified
        if profile is not None and profile == SqliteProfile.SPEED:
            # Use memory mapped IO: much faster for calculations
            # (max 30GB)
            sql = "PRAGMA mmap_size=30000000000;"
            conn.execute(sql)

            # These options don't really make a difference on windows, but
            # it doesn't hurt and maybe on other platforms...
            for databasename in database_names:
                sql = f"PRAGMA {databasename}.journal_mode=OFF;"
                conn.execute(sql)

                # These pragma's increase speed
                sql = f"PRAGMA {databasename}.locking_mode=EXCLUSIVE;"
                conn.execute(sql)
                sql = f"PRAGMA {databasename}.synchronous=OFF;"
                conn.execute(sql)

    except Exception as ex:  # pragma: no cover
        raise RuntimeError(f"Error executing {sql}: {ex}") from ex


def load_spatialite(conn: sqlite3.Connection, enable_gpkg_mode: bool) -> None:
    """Load mod_spatialite for an existing sqlite connection.

    Args:
        conn (sqlite3.Connection): Sqlite connection
        enable_gpkg_mode (bool): True if the database is a GeoPackage and GeoPackage
            mode should be activated.
    """
    conn.enable_load_extension(True)
    try:
        conn.load_extension("mod_spatialite")
    except Exception as ex:  # pragma: no cover
        raise MissingRuntimeDependencyError(
            "Error trying to load mod_spatialite."
        ) from ex

    if enable_gpkg_mode:
        gpkg_mode_failed = False
        try:
            sql = "SELECT EnableGpkgMode();"
            conn.execute(sql)

            # Verify if GeoPackage mode was enabled successfully
            sql = "SELECT GetGpkgMode();"
            result = conn.execute(sql).fetchone()
            gpkg_mode_failed = result is None or result[0] != 1
            if gpkg_mode_failed:
                raise RuntimeError("Failed to enable GPKG mode in mod_spatialite.")

        except Exception as ex:  # pragma: no cover
            if gpkg_mode_failed:
                raise
            raise RuntimeError(
                "Error trying to enable GPKG mode in mod_spatialite."
            ) from ex

    # Register custom functions
    # conn.create_function(
    #     "GFO_Difference_Collection",
    #     -1,
    #     sqlite_userdefined.gfo_difference_collection,
    #     deterministic=True,
    # )

    conn.create_function(
        "GFO_ReducePrecision",
        -1,
        _sqlite_userdefined.gfo_reduceprecision,
        deterministic=True,
    )

    # conn.create_function(
    #     "GFO_Split", -1, sqlite_userdefined.gfo_split, deterministic=True
    # )

    # conn.create_function(
    #     "GFO_Subdivide", -1, sqlite_userdefined.gfo_subdivide, deterministic=True
    # )

    # Register custom aggregate function
    # conn.create_aggregate("GFO_Difference_Agg", 3, userdefined.DifferenceAgg)



"""Module with compatibility and dependency availability checks."""

import sys

import geopandas as gpd
import pandas as pd
import pyogrio
import shapely
from osgeo import gdal
from packaging import version

from geofileops.util import _sqlite_util

gdal.UseExceptions()

# Determine the versions of the runtime dependencies
# gdal.__version__ includes a "dev-..." suffix for master/development versions. This
# must be dropped for the version checks here.
GDAL_BASE_VERSION = version.parse(gdal.__version__.split("-")[0]).base_version
GDAL_GTE_39 = version.parse(GDAL_BASE_VERSION) >= version.parse("3.9")
GDAL_GTE_3101 = version.parse(GDAL_BASE_VERSION) >= version.parse("3.10.1")
GDAL_GTE_311 = version.parse(GDAL_BASE_VERSION) >= version.parse("3.11")
GDAL_GTE_3114 = version.parse(GDAL_BASE_VERSION) >= version.parse("3.11.4")

GEOPANDAS_GTE_10 = version.parse(gpd.__version__) >= version.parse("1.0")
GEOPANDAS_110 = version.parse(gpd.__version__) == version.parse("1.1.0")

GEOS_GTE_313 = version.parse(shapely.geos_version_string) >= version.parse("3.13")

PANDAS_GTE_20 = version.parse(pd.__version__) >= version.parse("2.0")
PANDAS_GTE_22 = version.parse(pd.__version__) >= version.parse("2.2")
PANDAS_GTE_30 = version.parse(pd.__version__) >= version.parse("3.0")

PYOGRIO_GTE_010 = version.parse(pyogrio.__version__) >= version.parse("0.10")
PYOGRIO_GTE_011 = version.parse(pyogrio.__version__) >= version.parse("0.11")
PYOGRIO_GTE_012 = version.parse(pyogrio.__version__) >= version.parse("0.12")

PYTHON_313 = sys.version_info >= (3, 13) and sys.version_info < (3, 14)
SHAPELY_GTE_20 = version.parse(shapely.__version__) >= version.parse("2")

sqlite3_spatialite_version_info = _sqlite_util.spatialite_version_info()
sqlite3_spatialite_version = sqlite3_spatialite_version_info["spatialite_version"]



"""Module with union_full geooperations."""

import logging
from datetime import datetime
from pathlib import Path
from typing import Literal, TypeAlias, get_args

import geofileops as gfo
from geofileops import GeometryType, LayerInfo, fileops
from geofileops.helpers import _general_helper
from geofileops.helpers._parameter_helper import validate_params_single_layer
from geofileops.util import _geoops_sql, _io_util, _ogr_sql_util
from geofileops.util._geoops_sql import (
    _subdivide_layer,
    delete_duplicate_geometries,
    difference,
    intersection,
    join_by_location,
    select,
)

UnionFullSelfTypes: TypeAlias = Literal["COLUMNS", "LISTS", "ROWS"]


def union_full_self(
    input_path: Path,
    output_path: Path,
    intersections_as: UnionFullSelfTypes,
    input_layer: str | LayerInfo | None = None,
    output_layer: str | None = None,
    columns: list[str] | None = None,
    explodecollections: bool = False,
    gridsize: float = 0.0,
    where_post: str | None = None,
    nb_parallel: int | None = None,
    batchsize: int = -1,
    subdivide_coords: int = 2000,
    force: bool = False,
    output_with_spatial_index: bool | None = None,
) -> None:
    # Because the calculations of the intermediate results will be towards temp files,
    # we need to do some additional init + checks here...
    if subdivide_coords < 0:
        raise ValueError("subdivide_coords < 0 is not allowed")
    if intersections_as not in get_args(UnionFullSelfTypes):
        raise ValueError(
            f"intersections_as should be one of {get_args(UnionFullSelfTypes)}"
        )

    operation_name = "union_full_self"
    logger = logging.getLogger(f"geofileops.{operation_name}")

    if _io_util.output_exists(path=output_path, remove_if_exists=force):
        return

    input_layer, output_layer = validate_params_single_layer(
        input_path=input_path,
        output_path=output_path,
        input_layer=input_layer,
        output_layer=output_layer,
        operation_name=operation_name,
    )

    # Determine output_geometrytype
    force_output_geometrytype = input_layer.geometrytype
    if explodecollections:
        force_output_geometrytype = force_output_geometrytype.to_singletype
    elif force_output_geometrytype is not GeometryType.POINT:
        # If explodecollections is False and the input type is not point, force the
        # output type to multi, because difference can cause eg. polygons to be split to
        # multipolygons.
        force_output_geometrytype = force_output_geometrytype.to_multitype

    start_time = datetime.now()
    with _general_helper.create_gfo_tmp_dir(operation_name) as tmp_dir:
        # Prepare the input files
        logger.info("Step 1 of 4: prepare input file")
        input_subdivided_path = _subdivide_layer(
            path=input_path,
            layer=input_layer,
            output_path=tmp_dir / "subdivided/input_layer.gpkg",
            subdivide_coords=subdivide_coords,
            nb_parallel=nb_parallel,
            batchsize=batchsize,
            operation_prefix=f"{operation_name}/",
            tmp_basedir=tmp_dir,
        )
        if input_subdivided_path is None:
            # Hardcoded optimization: root means that no subdivide was needed
            input_subdivided_path = Path("/")

        # Loop until all intersections are gone...
        logger.info("Step 2 of 4: create a 'flat union' layer without intersections")
        non_intersecting_path = tmp_dir / "union_self_non_intersecting.gpkg"
        if intersections_as == "COLUMNS":
            columns_pass = list(input_layer.columns) if columns is None else columns
        else:
            columns_pass = []

        pass_id = 1
        input_path_pass = input_path
        input_layer_pass: str | LayerInfo = input_layer
        input_subdivided_cur_path: Path | None = input_subdivided_path
        while True:
            logger.info(f"  -> create 'flat union' layer, pass {pass_id}")
            if pass_id == 1:
                # In the first pass, include the original fid column
                input1_columns = columns_pass
                input2_columns = input1_columns
                input1_columns_prefix = f"i{pass_id}_"
                input2_columns_prefix = f"i{pass_id + 1}_"

            elif pass_id == 2:
                # In the 2nd pass, we need to include the original fid column
                input1_columns = None  # all columns
                input2_columns = [
                    f"i{pass_id}_{col}" for col in columns_pass
                ]  # only the "right" columns
                input1_columns_prefix = ""
                input2_columns_prefix = f"i{pass_id + 1}_"

                # From the 2nd pass on, we cannot use the subdivided input anymore
                input_subdivided_cur_path = None

                # From the 2nd pass on, input_pass_layer becomes the output layer
                input_layer_pass = output_layer

            else:
                # In the next passes, no use to keep columns anymore
                input1_columns = []
                input2_columns = []

            # Parts of geometries that don't intersect in the input layer are ready for
            # the output.
            # Remark: subdivide_coords=0, because the input is already subdivided.
            diff_output_path = tmp_dir / f"diff_output_{pass_id}.gpkg"
            difference(
                input1_path=input_path_pass,
                input2_path=input_path_pass,
                output_path=diff_output_path,
                overlay_self=True,
                input1_layer=input_layer_pass,
                input1_columns=input1_columns,
                input_columns_prefix=input1_columns_prefix,
                input2_layer=input_layer_pass,
                output_layer=output_layer,
                explodecollections=explodecollections,
                gridsize=gridsize,
                where_post=where_post,
                nb_parallel=nb_parallel,
                batchsize=batchsize,
                subdivide_coords=0,
                force=force,
                output_with_spatial_index=False,
                operation_prefix=f"{operation_name}/",
                tmp_basedir=tmp_dir,
                input1_subdivided_path=input_subdivided_cur_path,
                input2_subdivided_path=input_subdivided_cur_path,
            )

            if pass_id == 1:
                # First pass, so we can just rename
                gfo.move(diff_output_path, non_intersecting_path)
            else:
                # Append
                fileops.copy_layer(
                    src=diff_output_path,
                    dst=non_intersecting_path,
                    src_layer=output_layer,
                    dst_layer=output_layer,
                    write_mode="append_add_fields",
                    force_output_geometrytype=force_output_geometrytype,
                )

            # Determine the parts of geometries in the input layer that intersect.
            # Remark: subdivide_coords=0, because the input is already subdivided.
            intersection_output_path = tmp_dir / f"intersection_output_{pass_id}.gpkg"
            intersection(
                input1_path=input_path_pass,
                input2_path=input_path_pass,
                output_path=intersection_output_path,
                overlay_self=True,
                include_duplicates=False,
                input1_layer=input_layer_pass,
                input1_columns=input1_columns,
                input1_columns_prefix=input1_columns_prefix,
                input2_layer=input_layer_pass,
                input2_columns=input2_columns,
                input2_columns_prefix=input2_columns_prefix,
                output_layer=output_layer,
                explodecollections=explodecollections,
                gridsize=gridsize,
                where_post=None,
                nb_parallel=nb_parallel,
                batchsize=batchsize,
                subdivide_coords=0,
                force=force,
                output_with_spatial_index=False,
                operation_prefix=f"{operation_name}/",
                tmp_basedir=tmp_dir,
                input1_subdivided_path=input_subdivided_cur_path,
                input2_subdivided_path=input_subdivided_cur_path,
            )

            # If the intersection output is empty, we are ready...
            inters_info = gfo.get_layerinfo(path=intersection_output_path)
            if inters_info.featurecount == 0:
                attributes_in_flat_union = True if pass_id <= 2 else False
                break

            # Delete duplicates from the intersections before starting next pass.
            deldups_path = tmp_dir / f"{intersection_output_path.stem}_no-dups.gpkg"
            delete_duplicate_geometries(
                input_path=intersection_output_path,
                output_path=deldups_path,
                input_layer=output_layer,
                output_layer=output_layer,
                columns=None,
                priority_column=None,
                priority_ascending=True,
                explodecollections=explodecollections,
                keep_empty_geoms=False,
                where_post=None,
                nb_parallel=nb_parallel,
                batchsize=batchsize,
                force=force,
                operation_prefix=f"{operation_name}/",
                tmp_basedir=tmp_dir,
            )
            intersection_output_path = deldups_path

            # Init for the next pass
            # if intersection_output_prev_path is not None:
            #    gfo.remove(intersection_output_prev_path)
            input_path_pass = intersection_output_path
            pass_id += 1

        logger.info("Step 3 of 4: combine the attributes with the 'flat union' layer")
        if intersections_as == "COLUMNS" and columns == []:
            # No output attribute columns needed -> ready
            output_tmp_path = non_intersecting_path

        elif intersections_as == "COLUMNS" and attributes_in_flat_union:
            # If the attributes are already in the result of the "flat union", not
            # needed to join them.
            output_tmp_path = non_intersecting_path

        else:
            # Join the "flat union" with the original input layer to:
            #   - add the attributes of the input layer
            #   - duplicate the polygon parts as many times as they overlap in the
            #     input layer
            union_multirow_path = tmp_dir / "union_multirow.gpkg"
            join_by_location(
                input1_path=non_intersecting_path,
                input2_path=input_path,
                output_path=union_multirow_path,
                spatial_relations_query="intersects is True and touches is False",
                input1_layer=output_layer,
                input2_layer=input_layer,
                output_layer=output_layer,
                input1_columns=["fid"],
                input2_columns=columns,
                input1_columns_prefix="union_",
                input2_columns_prefix="",
                explodecollections=explodecollections,
                gridsize=gridsize,
                where_post=where_post,
                nb_parallel=nb_parallel,
                batchsize=batchsize,
                force=False,
                output_with_spatial_index=False,
                operation_prefix=f"{operation_name}/",
                tmp_basedir=tmp_dir,
            )
            output_tmp_path = union_multirow_path

            if intersections_as in ("COLUMNS", "LISTS"):
                columns_local = (
                    list(input_layer.columns) if columns is None else columns
                )
                sql_stmt = _get_union_full_attr_sql_stmt(
                    union_multirow_path=output_tmp_path,
                    intersections_as=intersections_as,
                    columns=columns_local,
                )
                select_output_path = tmp_dir / "union_with_attributes.gpkg"
                select(
                    input_path=output_tmp_path,
                    output_path=select_output_path,
                    sql_stmt=sql_stmt,
                    input_layer=output_layer,
                    output_layer=output_layer,
                    preserve_fid=False,
                    explodecollections=explodecollections,
                    nb_parallel=nb_parallel,
                    batchsize=batchsize,
                    operation_prefix=f"{operation_name}/",
                    batch_filter_column="union_fid",
                    tmp_basedir=tmp_dir,
                )
                output_tmp_path = select_output_path

        # Convert or add spatial index
        logger.info("Step 4 of 4: finalize")

        _geoops_sql._finalize_output(
            output_tmp_path=output_tmp_path,
            output_path=output_path,
            output_layer=output_layer,
            output_with_spatial_index=output_with_spatial_index,
        )

    logger.info(f"Ready, full {operation_name} took {datetime.now() - start_time}")


def _get_union_full_attr_sql_stmt(
    union_multirow_path: Path,
    intersections_as: UnionFullSelfTypes | str,
    columns: list[str],
) -> str:
    """Create a sql statement to aggregate attributes based on a multi-row-union file.

    The file should include a "union_fid" column to group on.

    Args:
        union_multirow_path (Path): path to the union file with multiple rows per
            intersection
        intersections_as (str): type of union_full_self
        columns (List[str]): list of columns to aggregate

    Returns:
        str: sql statement
    """
    if intersections_as == "LISTS":
        if len(columns) > 0:
            columns_list = []
            for col in columns:
                if col.lower() == "fid":
                    # The fid column will be aliased already in union_multirow_path!
                    alias = _ogr_sql_util.get_unique_fid_alias(col, columns)
                    column_str = f'json_group_array("{alias}") AS "{alias}"'
                else:
                    column_str = f'json_group_array("{col}") AS "{col}"'
                columns_list.append(column_str)

            columns_str = f", {', '.join(columns_list)}"

        else:
            columns_str = ""

        # An index on union_fid does not speed this up/decrease memory usage
        sql_stmt = f"""
            SELECT layer.{{geometrycolumn}}
                  ,COUNT(*) AS nb_intersecting
                  {columns_str}
              FROM "{{input_layer}}" layer
             WHERE 1=1
               {{batch_filter}}
             GROUP BY union_fid
        """

    elif intersections_as == "COLUMNS":
        # An index on union_fid does not speed up/decrease memory usage for
        # following queries
        if len(columns) > 0:
            # First determine the maximum number of intersections
            sql_stmt_max = """
                SELECT MAX(counts.count) AS max_intersections
                  FROM  ( SELECT COUNT(*) AS count
                            FROM "{input_layer}" layer
                           GROUP BY union_fid
                        ) counts
            """
            df = fileops.read_file(union_multirow_path, sql_stmt=sql_stmt_max)
            max_intersections = df.iloc[0]["max_intersections"]

            # Now create the columns for in the select
            columns_list = []
            for i_id in range(1, max_intersections + 1):
                for input_col in columns:
                    if input_col.lower() == "fid":
                        input_col = "fid_1"
                        col_alias = f"i{i_id}_fid"
                    else:
                        col_alias = f"i{i_id}_{input_col}"
                    columns_list.append(
                        f'MIN(CASE WHEN rn = {i_id + 1} THEN "{input_col}" END'
                        f') AS "{col_alias}"'
                    )

            columns_str = f", {', '.join(columns_list)}"

        else:
            columns_str = ""

        sql_stmt = f"""
            SELECT data.{{geometrycolumn}}
                  {columns_str}
              FROM (
                SELECT layer.{{geometrycolumn}}
                      ,layer.union_fid
                      {{columns_to_select_str}}
                      ,row_number() OVER (PARTITION BY union_fid) AS rn
                  FROM "{{input_layer}}" layer
                 WHERE 1=1
                   {{batch_filter}}
                ) data
             WHERE 1=1
             GROUP BY data.union_fid
        """

    else:
        raise ValueError(f"Unsupported union_type: {intersections_as}")

    return sql_stmt
